"""Configuration for ACLs."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import netaddr
from os_ken.ofproto import ether

from faucet import valve_of
from faucet import valve_acl
from faucet.valve_of import MATCH_FIELDS, OLD_MATCH_FIELDS
from faucet.conf import Conf, test_config_condition, InvalidConfigError
from faucet.valve_table import wildcard_table


class ACL(Conf):
    """Contains the state for an ACL, including the configuration.

ACL Config

ACLs are configured under the 'acls' configuration block. The acls block
contains a dictionary of individual acls each keyed by its name.

Each acl contains a list of rules, a packet will have the first matching rule
applied to it.

Each rule is a dictionary containing the single key 'rule' with the value the
matches and actions for the rule.

The matches are key/values based on the ryu RESTFul API.
The key 'actions' contains a dictionary with keys/values as follows:

 * allow (int): if 1 allow the packet to continue through the Faucet \
       pipeline, if 0 drop the packet.
 * force_port_vlan (int): if 1, do not verify the VLAN/port association \
       for this packet and override any VLAN ACL on the forced VLAN.
 * meter (str): meter to apply to the packet
 * output (dict): used to output a packet directly. details below.
 * cookie (int): set flow cookie to this value on this flow

The output action contains a dictionary with the following elements:

 * tunnel (dict): the tunnel formation, creates a tunnel from the applied port(s) \
       to the specified destination
 * port (int or string): the port to output the packet to
 * ports (list): a list of the ports (int or string) to output the packet to
 * set_fields (list): a list of fields to set with values
 * pop_vlans: (int): pop the packet vlan before outputting
 * vlan_vid: (int): push the vlan vid on the packet when outputting
 * vlan_vids: (list): push the list of vlans on the packet when outputting, with option eth_type
 * swap_vid (int): rewrite the vlan vid of the packet when outputting
 * failover (dict): Output with a failover port (experimental)
"""

    defaults = {
        'rules': None,
        'exact_match': False,
        'dot1x_assigned': False,
    }
    defaults_types = {
        'rules': list,
        'exact_match': bool,
        'dot1x_assigned': bool,
    }
    rule_types = {
        'cookie': int,
        'actions': dict,
        'description': str,
    }
    actions_types = {
        'meter': str,
        'mirror': (str, int),
        'output': (dict, list),
        'allow': int,
        'force_port_vlan': int,
        'ct': dict,
    }
    output_actions_types = {
        'tunnel': dict,
        'port': (str, int),
        'ports': list,
        'failover': dict,
        'set_fields': list,
        'pop_vlans': int,
        'swap_vid': int,
        'vlan_vid': int,
        'vlan_vids': list,
    }
    ct_action_types = {
        'flags': int,
        'alg': int,
        'table': int,
        'zone': int,
        'zone_src': int,
        'clear': bool,
        'nat': dict,
    }
    ct_action_nat_types = {
        'flags': int,
        'range_ipv4_min': str,
        'range_ipv4_max': str,
        'range_ipv6_min': str,
        'range_ipv6_max': str,
        'range_proto_min': int,
        'range_proto_max': int
    }
    tunnel_types = {
        'type': (str, None),
        'tunnel_id': (str, int, None),
        'dp': str,
        'port': (str, int, None),
        'exit_instructions': (list, None),
        'maintain_encapsulation': bool,
        'bi_directional': bool,
        'reverse': bool,
    }

    mutable_attrs = frozenset(['tunnel_sources'])

    def __init__(self, _id, dp_id, conf):
        self.rules = []
        self.exact_match = None
        self.dot1x_assigned = None
        self.meter = False
        self.matches = {}
        self.set_fields = set()
        self._ports_resolved = False

        # Tunnel info maintains the tunnel output information for each tunnel rule
        self.tunnel_dests = {}
        # Tunnel sources is a list of the sources in the network for this ACL
        self.tunnel_sources = {}
        # Tunnel rules is the rules for each tunnel in the ACL for each source
        self.dyn_tunnel_rules = {}
        self.dyn_reverse_tunnel_rules = {}

        for match_fields in (MATCH_FIELDS, OLD_MATCH_FIELDS):
            self.rule_types.update({match: (str, int) for match in match_fields})
        conf = copy.deepcopy(conf)
        if isinstance(conf, dict):
            rules = conf.get('rules', [])
        elif isinstance(conf, list):
            rules = conf
            conf = {}
        else:
            raise InvalidConfigError(
                f'ACL conf is an invalid type {_id}')
        conf['rules'] = []
        for rule in rules:
            normalized_rule = rule
            if isinstance(rule, dict):
                normalized_rule = rule.get('rule', rule)
                if normalized_rule is None:
                    normalized_rule = {k: v for k, v in rule.items() if v is not None}
            test_config_condition(not isinstance(normalized_rule, dict), (
                f'ACL rule is {type(normalized_rule)} not {dict} ({rules})'))
            conf['rules'].append(normalized_rule)
        super().__init__(_id, dp_id, conf)

    def finalize(self):
        self._ports_resolved = True
        super().finalize()

    def check_config(self):
        test_config_condition(
            not self.rules, f'no rules found for ACL {self._id}')
        for rule in self.rules:
            self._check_conf_types(rule, self.rule_types)
            for rule_field, rule_conf in rule.items():
                if rule_field == 'cookie':
                    test_config_condition(
                        rule_conf < 0 or rule_conf > 2**16,
                        'rule cookie value must be 0-2**16')
                elif rule_field == 'actions':
                    test_config_condition(
                        not rule_conf,
                        f'Missing rule actions in ACL {self._id}')
                    self._check_conf_types(rule_conf, self.actions_types)
                    for action_name, action_conf in rule_conf.items():
                        if action_name == 'output':
                            if isinstance(action_conf, (list, tuple)):
                                # New ordered format
                                for subconf in action_conf:
                                    # Make sure only one specified action per list element
                                    test_config_condition(
                                        len(subconf) > 1,
                                        'ACL ordered output must have only one action per element')
                                    # Ensure correct action format
                                    self._check_conf_types(subconf, self.output_actions_types)
                            else:
                                # Old format
                                self._check_conf_types(
                                    action_conf, self.output_actions_types)
                        elif action_name == 'ct':
                            self._check_conf_types(action_conf, self.ct_action_types)
                            # if clear set, make sure nothing else is
                            if 'clear' in action_conf and action_conf['clear']:
                                test_config_condition(
                                    len(action_conf) != 1,
                                    "no other parameters can be set when 'clear' set on "
                                    "conntrack ACL")
                            else:
                                test_config_condition(
                                    'table' not in action_conf,
                                    "required parameter 'table' not set for conntrack ACL")
                                test_config_condition(
                                    'zone' not in action_conf,
                                    "required parameter 'zone' not set for conntrack ACL")
                            if 'nat' in action_conf:
                                self._check_conf_types(action_conf['nat'], self.ct_action_nat_types)

    def build(self, meters, vid, port_num):
        """Check that ACL can be built from config."""

        self.matches = {}
        self.set_fields = set()
        self.meter = False
        if self.rules:
            try:
                ofmsgs = valve_acl.build_acl_ofmsgs(
                    [self], wildcard_table,
                    [valve_of.goto_table(wildcard_table)],
                    [valve_of.goto_table(wildcard_table)],
                    2**16 - 1, meters, self.exact_match,
                    vlan_vid=vid, port_num=port_num)
            except (netaddr.core.AddrFormatError, KeyError, ValueError) as err:
                raise InvalidConfigError from err
            test_config_condition(not ofmsgs, 'OF messages is empty')
            for ofmsg in ofmsgs:
                try:
                    valve_of.verify_flowmod(ofmsg)
                except (KeyError, ValueError) as err:
                    raise InvalidConfigError from err
                except Exception as err:
                    raise err
                if valve_of.is_flowmod(ofmsg):
                    apply_actions = []
                    for inst in ofmsg.instructions:
                        if valve_of.is_apply_actions(inst):
                            apply_actions.extend(inst.actions)
                        elif valve_of.is_meter(inst):
                            self.meter = True
                    for action in apply_actions:
                        if valve_of.is_set_field(action):
                            self.set_fields.add(action.key)
                    for match, value in ofmsg.match.items():
                        has_mask = isinstance(value, (tuple, list))
                        if has_mask or match not in self.matches:
                            self.matches[match] = has_mask
        for tunnel_rules in self.tunnel_dests.values():
            if 'exit_instructions' in tunnel_rules:
                exit_inst = tunnel_rules['exit_instructions']
                try:
                    ofmsgs = valve_acl.build_tunnel_ofmsgs(
                        exit_inst, wildcard_table, 1)
                except (netaddr.core.AddrFormatError, KeyError, ValueError) as err:
                    raise InvalidConfigError from err
                test_config_condition(not ofmsgs, 'OF messages is empty')
                for ofmsg in ofmsgs:
                    try:
                        valve_of.verify_flowmod(ofmsg)
                    except (KeyError, ValueError) as err:
                        raise InvalidConfigError from err
                    except Exception as err:
                        raise err
                    if valve_of.is_flowmod(ofmsg):
                        apply_actions = []
                        for inst in ofmsg.instructions:
                            if valve_of.is_apply_actions(inst):
                                apply_actions.extend(inst.actions)
                            elif valve_of.is_meter(inst):
                                self.meter = True
                        for action in apply_actions:
                            if valve_of.is_set_field(action):
                                self.set_fields.add(action.key)
                        for match, value in ofmsg.match.items():
                            has_mask = isinstance(value, (tuple, list))
                            if has_mask or match not in self.matches:
                                self.matches[match] = has_mask
        return (self.matches, self.set_fields, self.meter)

    def get_meters(self):
        """Yield meters for each rule in ACL"""
        for rule in self.rules:
            if 'actions' not in rule or 'meter' not in rule['actions']:
                continue
            yield rule['actions']['meter']

    def get_mirror_destinations(self):
        """Yield mirror destinations for each rule in ACL"""
        for rule in self.rules:
            if 'actions' not in rule or 'mirror' not in rule['actions']:
                continue
            yield rule['actions']['mirror']

    def _resolve_ordered_output_ports(self, output_list, resolve_port_cb, resolve_tunnel_objects):
        """Resolve output actions in the ordered list format"""
        result = []
        for action in output_list:
            for key, value in action.items():
                if key == 'tunnel':
                    tunnel = value
                    # Fetch tunnel items from the tunnel output dict
                    test_config_condition(
                        'dp' not in tunnel,
                        f'ACL ({self._id}) tunnel DP not defined')
                    tunnel_dp = tunnel['dp']
                    tunnel_port = tunnel.get('port', None)
                    tunnel_id = tunnel.get('tunnel_id', None)
                    tunnel_type = tunnel.get('type', 'vlan')
                    tunnel_exit_instructions = tunnel.get('exit_instructions', [])
                    tunnel_direction = tunnel.get('bi_directional', False)
                    tunnel_maintain = tunnel.get('maintain_encapsulation', False)
                    tunnel_reverse = tunnel.get('reverse', False)
                    test_config_condition(
                        tunnel_reverse and tunnel_direction,
                        (f'Tunnel ACL {self._id} cannot contain values for the fields'
                         '`bi_directional` and `reverse` at the same time'))
                    # Resolve the tunnel items
                    dst_dp, dst_port, tunnel_id = resolve_tunnel_objects(
                        tunnel_dp, tunnel_port, tunnel_id)
                    # Compile the tunnel into an easy-access dictionary
                    tunnel_dict = {
                        'dst_dp': dst_dp,
                        'dst_port': dst_port,
                        'tunnel_id': tunnel_id,
                        'type': tunnel_type,
                        'exit_instructions': tunnel_exit_instructions,
                        'bi_directional': tunnel_direction,
                        'maintain_encapsulation': tunnel_maintain,
                        'reverse': tunnel_reverse,
                    }
                    self.tunnel_dests[tunnel_id] = tunnel_dict
                    result.append({key: tunnel_id})
                elif key == 'port':
                    port_name = value
                    port = resolve_port_cb(port_name)
                    test_config_condition(
                        not port,
                        f'ACL ({self._id}) output port undefined in DP: {self.dp_id}')
                    result.append({key: port})
                elif key == 'ports':
                    resolved_ports = [
                        resolve_port_cb(p) for p in value]
                    test_config_condition(
                        None in resolved_ports,
                        f'ACL ({self._id}) output port(s) not defined in DP: {self.dp_id}')
                    result.append({key: resolved_ports})
                elif key == 'failover':
                    failover = value
                    test_config_condition(not isinstance(failover, dict), (
                        'failover is not a dictionary'))
                    failover_dict = {}
                    for failover_name, failover_values in failover.items():
                        if failover_name == 'ports':
                            resolved_ports = [
                                resolve_port_cb(p) for p in failover_values]
                            test_config_condition(
                                None in resolved_ports,
                                f'ACL ({self._id}) failover port(s) not defined in DP: {self.dp_id}')
                            failover_dict[failover_name] = resolved_ports
                        else:
                            failover_dict[failover_name] = failover_values
                    result.append({key: failover_dict})
                else:
                    result.append(action)
        return result

    def _resolve_output_ports(self, action_conf, resolve_port_cb, resolve_tunnel_objects):
        """Resolve the values for output actions in the ACL"""
        if isinstance(action_conf, (list, tuple)):
            return self._resolve_ordered_output_ports(
                action_conf, resolve_port_cb, resolve_tunnel_objects)
        result = {}
        test_config_condition(
            'vlan_vid' in action_conf and 'vlan_vids' in action_conf,
            f'ACL {self._id} has both vlan_vid and vlan_vids defined')
        test_config_condition(
            'port' in action_conf and 'ports' in action_conf,
            f'ACL {self._id} has both port and ports defined')
        for output_action, output_action_values in action_conf.items():
            if output_action == 'tunnel':
                tunnel = output_action_values
                # Fetch tunnel items from the tunnel output dict
                test_config_condition(
                    'dp' not in tunnel,
                    f'ACL ({self._id}) tunnel DP not defined')
                tunnel_dp = tunnel['dp']
                tunnel_port = tunnel.get('port', None)
                tunnel_id = tunnel.get('tunnel_id', None)
                tunnel_type = tunnel.get('type', 'vlan')
                tunnel_exit_instructions = tunnel.get('exit_instructions', [])
                tunnel_direction = tunnel.get('bi_directional', False)
                tunnel_maintain = tunnel.get('maintain_encapsulation', False)
                tunnel_reverse = tunnel.get('reverse', False)
                test_config_condition(
                    tunnel_reverse and tunnel_direction,
                    (f'Tunnel ACL {self._id} cannot contain values for the fields'
                     '`bi_directional` and `reverse` at the same time')
                )
                # Resolve the tunnel items
                dst_dp, dst_port, tunnel_id = resolve_tunnel_objects(
                    tunnel_dp, tunnel_port, tunnel_id)
                # Compile the tunnel into an easy-access dictionary
                tunnel_dict = {
                    'dst_dp': dst_dp,
                    'dst_port': dst_port,
                    'tunnel_id': tunnel_id,
                    'type': tunnel_type,
                    'exit_instructions': tunnel_exit_instructions,
                    'bi_directional': tunnel_direction,
                    'maintain_encapsulation': tunnel_maintain,
                    'reverse': tunnel_reverse,
                }
                self.tunnel_dests[tunnel_id] = tunnel_dict
                result[output_action] = tunnel_id
            elif output_action == 'port':
                port_name = output_action_values
                port = resolve_port_cb(port_name)
                test_config_condition(
                    not port,
                    (f'ACL ({self._id}) output port undefined in DP: {self.dp_id}')
                )
                result[output_action] = port
            elif output_action == 'ports':
                resolved_ports = [
                    resolve_port_cb(p) for p in output_action_values]
                test_config_condition(
                    None in resolved_ports,
                    (f'ACL ({self._id}) output port(s) not defined in DP: {self.dp_id}')
                )
                result[output_action] = resolved_ports
            elif output_action == 'failover':
                failover = output_action_values
                test_config_condition(not isinstance(failover, dict), (
                    'failover is not a dictionary'))
                result[output_action] = {}
                for failover_name, failover_values in failover.items():
                    if failover_name == 'ports':
                        resolved_ports = [
                            resolve_port_cb(p) for p in failover_values]
                        test_config_condition(
                            None in resolved_ports,
                            (f'ACL ({self._id}) failover port(s) not defined in DP: {self.dp_id}')
                        )
                        result[output_action][failover_name] = resolved_ports
                    else:
                        result[output_action][failover_name] = failover_values
            else:
                result[output_action] = output_action_values
        return result

    def resolve_ports(self, resolve_port_cb, resolve_tunnel_objects):
        """Resolve the values for the actions of an ACL"""
        if self._ports_resolved:
            return
        for rule_conf in self.rules:
            if 'actions' in rule_conf:
                actions_conf = rule_conf['actions']
                resolved_actions = {}
                test_config_condition(not isinstance(actions_conf, dict), (
                    'actions value is not a dictionary'))
                for action_name, action_conf in actions_conf.items():
                    if action_name == 'mirror':
                        resolved_port = resolve_port_cb(action_conf)
                        test_config_condition(
                            resolved_port is None,
                            (f'ACL ({self._id}) mirror port is not defined in DP: {self.dp_id}')
                        )
                        resolved_actions[action_name] = resolved_port
                    elif action_name == 'output':
                        resolved_action = self._resolve_output_ports(
                            action_conf, resolve_port_cb, resolve_tunnel_objects)
                        resolved_actions[action_name] = resolved_action
                    else:
                        resolved_actions[action_name] = action_conf
                rule_conf['actions'] = resolved_actions
        self._ports_resolved = True

    def requires_reverse_tunnel(self, tunnel_id):
        """Returns true if the tunnel requires a reverse pathway"""
        return self.tunnel_dests[tunnel_id]['bi_directional']

    def get_num_tunnels(self):
        """Returns the number of tunnels specified in the ACL"""
        num_tunnels = 0
        for rule_conf in self.rules:
            if self.does_rule_contain_tunnel(rule_conf):
                output_conf = rule_conf['actions']['output']
                if isinstance(output_conf, list):
                    for action in output_conf:
                        for key in action:
                            if key == 'tunnel':
                                num_tunnels += 1
                else:
                    if 'tunnel' in output_conf:
                        num_tunnels += 1
        return num_tunnels

    def get_tunnel_rules(self, tunnel_id):
        """Return the list of rules that apply a specific tunnel ID"""
        rules = []
        for rule_conf in self.rules:
            if self.does_rule_contain_tunnel(rule_conf):
                output_conf = rule_conf['actions']['output']
                if isinstance(output_conf, (list, tuple)):
                    for action in output_conf:
                        for key, value in action.items():
                            if key == 'tunnel' and value == tunnel_id:
                                rules.append(rule_conf)
                                continue
                else:
                    if output_conf['tunnel'] == tunnel_id:
                        rules.append(rule_conf)
        return rules

    @staticmethod
    def does_rule_contain_tunnel(rule_conf):
        """Return true if the ACL rule contains a tunnel"""
        if 'actions' in rule_conf:
            if 'output' in rule_conf['actions']:
                output_conf = rule_conf['actions']['output']
                if isinstance(output_conf, (list, tuple)):
                    for action in output_conf:
                        for key in action:
                            if key == 'tunnel':
                                return True
                else:
                    if 'tunnel' in output_conf:
                        return True
        return False

    def is_tunnel_acl(self):
        """Return true if the ACL contains a tunnel"""
        if self.tunnel_dests:
            return True
        for rule_conf in self.rules:
            if self.does_rule_contain_tunnel(rule_conf):
                return True
        return False

    @staticmethod
    def _tunnel_source_id(source):
        """Return ID for a tunnel source."""
        return tuple(sorted(source.items()))

    def add_tunnel_source(self, dp_name, port, reverse=False, bi_directional=False):
        """Add a source dp/port pair for the tunnel ACL"""
        source = {'dp': dp_name, 'port': port, 'reverse': reverse, 'bi_directional': bi_directional}
        source_id = self._tunnel_source_id(source)
        self.tunnel_sources[source_id] = source
        for _id in self.tunnel_dests:
            self.dyn_tunnel_rules.setdefault(_id, {})
            self.dyn_reverse_tunnel_rules.setdefault(_id, {})

    def verify_tunnel_rules(self):
        """Make sure that matches & set fields are configured correctly to handle tunnels"""
        if 'eth_type' not in self.matches:
            self.matches['eth_type'] = False
        if 'in_port' not in self.matches:
            self.matches['in_port'] = False
        if 'vlan_vid' not in self.matches:
            self.matches['vlan_vid'] = False
        if 'vlan_vid' not in self.set_fields:
            self.set_fields.add('vlan_vid')
        if 'vlan_pcp' not in self.matches:
            self.matches['vlan_pcp'] = False
        if 'vlan_pcp' not in self.set_fields:
            self.set_fields.add('vlan_pcp')

    def update_reverse_tunnel_rules(self, curr_dp, source_id, tunnel_id, out_port, output_table):
        """Update the tunnel rulelist for when the output port has changed (reverse direction)"""
        if not self.requires_reverse_tunnel(tunnel_id):
            return False
        dst_dp = self.tunnel_sources[source_id]['dp']
        src_dp = self.tunnel_dests[tunnel_id]['dst_dp']
        prev_list = self.dyn_reverse_tunnel_rules[tunnel_id].get(source_id, [])
        new_list = []
        if curr_dp == src_dp and curr_dp != dst_dp:
            # SRC DP: vlan_vid, vlan_pcp, actions=[out_port]
            # NOTE: For the bi_directional reverse tunnel, we assume that
            #       the packet already has the required encapsulation
            new_list = [{'port': out_port}]
        elif curr_dp == dst_dp and curr_dp != src_dp:
            # DST DP: vlan_vid, vlan_pcp, actions=[pop_vlans, output]
            new_list = [{'pop_vlans': 1}]
            if out_port is None:
                # DP dest tunnel, so we fall through into the eth_dst output table
                new_list.append({'goto': output_table.table_id})
            else:
                # Tunnel has port specified, so output to destination
                new_list.append({'port': out_port})
        elif curr_dp == src_dp and curr_dp == dst_dp:
            # SINGLE DP: actions=[pop_vlans, out_port]
            new_list = [{'pop_vlans': 1}]
            if out_port is None:
                # DP dest tunnel, so we fall through into the eth_dst output table
                new_list.extend([{'goto': output_table.table_id}])
            else:
                # Tunnel has port specified, so output to destination
                new_list.extend([{'port': out_port}])
        else:
            # TRANSIT DP: vlan_vid, vlan_pcp, actions=[output]
            new_list = [{'port': out_port}]
        if new_list != prev_list:
            self.dyn_reverse_tunnel_rules[tunnel_id][source_id] = new_list
            return True
        return True

    def update_source_tunnel_rules(self, curr_dp, source_id, tunnel_id, out_port, output_table):
        """Update the tunnel rulelist for when the output port has changed"""
        src_dp = self.tunnel_sources[source_id]['dp']
        dst_dp = self.tunnel_dests[tunnel_id]['dst_dp']
        prev_list = self.dyn_tunnel_rules[tunnel_id].get(source_id, [])
        new_list = []
        pcp_flag = valve_of.PCP_TUNNEL_FLAG
        if self.tunnel_dests[tunnel_id]['reverse']:
            pcp_flag = valve_of.PCP_TUNNEL_REVERSE_DIRECTION_FLAG
        if curr_dp == src_dp and curr_dp != dst_dp:
            # SRC DP: in_port, actions=[push_vlan, output, pop_vlans]
            # Ideally, we would be able to detect if the tunnel has an `allow` action clause.
            #   However, this is difficult as a single ACL can have multiple rules using the same
            #   tunnel, but with one instance requiring the `allow` clause and another, not.
            # This means it is easier to always append the `pop_vlans` in assumption that the
            #   `allow` action does exist, and then optimize/reduce the redundant rules before
            #   outputting the flowrule.
            # We also set the tunnel VLAN header with a PCP value indicating that we are in
            #   the tunnel, which will save the VLANs from being reserved.
            new_list = [
                {'vlan_vids': [{'vid': tunnel_id, 'eth_type': ether.ETH_TYPE_8021Q}]},
                {'set_fields': [{'vlan_pcp': pcp_flag}]},
                {'port': out_port},
                {'pop_vlans': 1}]
        elif curr_dp == dst_dp and curr_dp != src_dp:
            # DST DP: in_port, vlan_vid, actions=[pop_vlan, additional_instructions, output]
            # If exit_instructions are applied, then we want to pop off the tunnel
            #   VLAN header, then apply the additional instructions, then output
            if self.tunnel_dests[tunnel_id]['maintain_encapsulation']:
                # We wish to maintain tunnel encapsulation before outputting
                #   So do not add the pop_vlans rule
                new_list = []
            else:
                new_list = [{'pop_vlans': 1}]
            exit_instructions = self.tunnel_dests[tunnel_id].get('exit_instructions', [])
            new_list.extend(copy.copy(list(exit_instructions)))
            if out_port is None:
                # DP dest tunnel, so we fall through into the eth_dst output table
                new_list.append({'goto': output_table.table_id})
            else:
                # Tunnel has port specified, so output to destination
                new_list.append({'port': out_port})
        elif curr_dp == src_dp and curr_dp == dst_dp:
            # SINGLE DP: in_port, actions=[additional_instructions, out_port]
            exit_instructions = self.tunnel_dests[tunnel_id].get('exit_instructions', [])
            new_list.extend(copy.copy(list(exit_instructions)))
            if self.tunnel_dests[tunnel_id].get('maintain_encapsulation', False):
                # Maintain encapsulation implies we want the tunnel VID on the packet,
                #   so ensure it is purposefully put onto the packet, even when
                #   there would originally be no need to push on a tunnel VID
                new_list.extend([
                    {'vlan_vids': [{'vid': tunnel_id, 'eth_type': ether.ETH_TYPE_8021Q}]},
                    {'set_fields': [{'vlan_pcp': pcp_flag}]}])
            if out_port is None:
                # DP dest tunnel, so we fall through into the eth_dst output table
                new_list.extend([{'goto': output_table.table_id}])
            else:
                # Tunnel has port specified, so output to destination
                new_list.extend([{'port': out_port}])
        else:
            # TRANSIT DP: in_port, vlan_vid, actions=[output]
            new_list = [{'port': out_port}]
        if new_list != prev_list:
            self.dyn_tunnel_rules[tunnel_id][source_id] = new_list
            return True
        return True


# NOTE: 802.1x steals the port ACL table.
PORT_ACL_8021X = ACL(
    'port_acl_8021x', 0,
    {'rules': [
        {'eth_type': 1, 'eth_src': '01:02:03:04:05:06', 'actions': {'output': {
            'port': valve_of.ofp.OFPP_LOCAL, 'set_fields': [
                {'eth_src': '01:02:03:04:05:06'}, {'eth_dst': '01:02:03:04:05:06'}]}}}]})
PORT_ACL_8021X.build({}, None, 1)

MAB_ACL_8021X = ACL(
    'mab_acl_8021x', 0,
    {'rules': [{
        'eth_type': valve_of.ether.ETH_TYPE_IP, 'eth_src': '01:02:03:04:05:06',
        'ip_proto': valve_of.inet.IPPROTO_UDP, 'udp_src': 68, 'udp_dst': 67,
        'actions': {'output': {'port': valve_of.ofp.OFPP_LOCAL}}}]})
MAB_ACL_8021X.build({}, None, 1)



"""OSKenApp shim between Ryu and Valve."""

# Copyright (C) 2013 Nippon Telegraph and Telephone Corporation.
# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=using-constant-test,wrong-import-order,wrong-import-position
import eventlet
if True:                # A trick to satisfy linting for E402
    eventlet.monkey_patch()

import time

from functools import partial

from os_ken.controller.handler import CONFIG_DISPATCHER
from os_ken.controller.handler import MAIN_DISPATCHER
from os_ken.controller.handler import set_ev_cls
from os_ken.controller import dpset
from os_ken.controller import event
from os_ken.controller import ofp_event
from os_ken.lib import hub

from faucet.valve_ryuapp import EventReconfigure, OSKenAppBase
from faucet.valve_util import dpid_log, kill_on_exception
from faucet import faucet_event
from faucet import faucet_bgp
from faucet import faucet_dot1x
from faucet import valves_manager
from faucet import faucet_metrics
from faucet import valve_of


EXPORT_RYU_CONFIGS = ['echo_request_interval', 'maximum_unreplied_echo_requests']


class EventFaucetMaintainStackRoot(event.EventBase):  # pylint: disable=too-few-public-methods
    """Event used to maintain stack root."""


class EventFaucetMetricUpdate(event.EventBase):  # pylint: disable=too-few-public-methods
    """Event used to trigger update of metrics."""


class EventFaucetResolveGateways(event.EventBase):  # pylint: disable=too-few-public-methods
    """Event used to trigger gateway re/resolution."""


class EventFaucetStateExpire(event.EventBase):  # pylint: disable=too-few-public-methods
    """Event used to trigger expiration of state in controller."""


class EventFaucetFastStateExpire(event.EventBase):  # pylint: disable=too-few-public-methods
    """Event used to trigger fast expiration of state in controller."""


class EventFaucetAdvertise(event.EventBase):  # pylint: disable=too-few-public-methods
    """Event used to trigger periodic network advertisements (eg IPv6 RAs)."""


class EventFaucetFastAdvertise(event.EventBase):  # pylint: disable=too-few-public-methods
    """Event used to trigger periodic fast network advertisements (eg LACP)."""


class EventFaucetEventSockHeartbeat(event.EventBase):  # pylint: disable=too-few-public-methods
    """Event used to trigger periodic events on event sock,
    causing it to raise an exception if conn is broken.
    """


class Faucet(OSKenAppBase):
    """A OSKenApp that implements an L2/L3 learning VLAN switch.

    Valve provides the switch implementation; this is a shim for the Ryu
    event handling framework to interface with Valve.
    """
    _CONTEXTS = {
        'dpset': dpset.DPSet,
    }
    _VALVE_SERVICES = {
        EventFaucetMetricUpdate: (None, 5),
        EventFaucetResolveGateways: ('resolve_gateways', 2),
        EventFaucetStateExpire: ('state_expire', 5),
        EventFaucetFastStateExpire: ('fast_state_expire', 2),
        EventFaucetAdvertise: ('advertise', 15),
        EventFaucetFastAdvertise: ('fast_advertise', 5),
    }
    logname = 'faucet'
    exc_logname = logname + '.exception'
    bgp = None
    notifier = None
    valves_manager = None
    event_socket_heartbeat_time = 0

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.prom_client = faucet_metrics.FaucetMetrics(reg=self._reg)
        self.bgp = faucet_bgp.FaucetBgp(
            self.logger, self.exc_logname, self.prom_client, self._send_flow_msgs)
        self.dot1x = faucet_dot1x.FaucetDot1x(
            self.logger, self.exc_logname, self.prom_client, self._send_flow_msgs)
        self.notifier = faucet_event.FaucetEventNotifier(
            self.get_setting('EVENT_SOCK'), self.prom_client, self.logger)
        self.valves_manager = valves_manager.ValvesManager(
            self.logname, self.logger, self.prom_client, self.notifier, self.bgp,
            self.dot1x, self.get_setting('CONFIG_AUTO_REVERT'), self._send_flow_msgs)
        self.thread_managers = (self.bgp, self.dot1x, self.prom_client, self.notifier)
        self.event_sock_hrtbeat_time = int(self.get_setting('EVENT_SOCK_HEARTBEAT') or 0)
        if self.event_sock_hrtbeat_time > 0:
            self._VALVE_SERVICES[EventFaucetEventSockHeartbeat] = ('event_sock_heartbeat',
                                                                   self.event_sock_hrtbeat_time)
        self.stack_root_state_update_time = int(
            self.get_setting('STACK_ROOT_STATE_UPDATE_TIME') or 0)
        if self.stack_root_state_update_time:
            self._VALVE_SERVICES[EventFaucetMaintainStackRoot] = (None,
                                                                  self.stack_root_state_update_time)

    @kill_on_exception(exc_logname)
    def _check_thread_exception(self):
        super()._check_thread_exception()

    def _export_ryu_config(self):
        for opt_name in EXPORT_RYU_CONFIGS:
            value = int(getattr(self.CONF, opt_name))
            config_labels = dict(param=opt_name)
            self.prom_client.ryu_config.labels(**config_labels).set(value)

    @kill_on_exception(exc_logname)
    def start(self):
        super().start()

        # Start Prometheus
        prom_port = int(self.get_setting('PROMETHEUS_PORT'))
        prom_addr = self.get_setting('PROMETHEUS_ADDR')
        self.prom_client.start(prom_port, prom_addr)
        self._export_ryu_config()

        # Start event notifier
        notifier_thread = self.notifier.start()
        if notifier_thread is not None:
            self.threads.append(notifier_thread)

        for service_event, service_pair in self._VALVE_SERVICES.items():
            name, interval = service_pair
            thread = hub.spawn(
                partial(self._thread_reschedule, service_event(), interval))
            thread.name = name
            self.threads.append(thread)

    def _delete_deconfigured_dp(self, deleted_dpid):
        self.logger.info(
            'Deleting de-configured %s', dpid_log(deleted_dpid))
        ryu_dp = self.dpset.get(deleted_dpid)
        if ryu_dp is not None:
            ryu_dp.close()

    @set_ev_cls(EventReconfigure, MAIN_DISPATCHER)
    @kill_on_exception(exc_logname)
    def reload_config(self, ryu_event):
        """Handle a request to reload configuration."""
        super().reload_config(ryu_event)
        self.valves_manager.request_reload_configs(
            time.time(), self.config_file, delete_dp=self._delete_deconfigured_dp)

    @kill_on_exception(exc_logname)
    def _send_flow_msgs(self, valve, flow_msgs, ryu_dp=None):
        """Send OpenFlow messages to a connected datapath.

        Args:
            Valve instance or None.
            flow_msgs (list): OpenFlow messages to send.
            ryu_dp: Override datapath from DPSet.
        """
        if ryu_dp is None:
            ryu_dp = self.dpset.get(valve.dp.dp_id)
        if not ryu_dp:
            valve.logger.error('send_flow_msgs: DP not up')
            return
        valve.send_flows(ryu_dp, flow_msgs, time.time())

    def _get_valve(self, ryu_event, require_running=False):
        """Get Valve instance to response to an event.

        Args:
            ryu_event (ryu.controller.event.Event): event
            require_running (bool): require DP to be running.
        Returns:
            valve, ryu_dp, msg: tuple of Nones, or datapath object, Ryu datapath, and msg (if any)
        """
        valve, ryu_dp, msg = self._get_datapath_obj(
            self.valves_manager.valves, ryu_event)
        if valve:
            if msg:
                valve.ofchannel_log([msg])
            if require_running and not valve.dp.dyn_running:
                valve = None
        return (valve, ryu_dp, msg)

    def _config_files_changed(self):
        return self.valves_manager.config_watcher.files_changed()

    @set_ev_cls(EventFaucetMetricUpdate, MAIN_DISPATCHER)
    @kill_on_exception(exc_logname)
    def metric_update(self, _):
        """Handle a request to update metrics in the controller."""
        self.valves_manager.update_metrics(time.time())

    @set_ev_cls(EventFaucetMaintainStackRoot, MAIN_DISPATCHER)
    @kill_on_exception(exc_logname)
    def _maintain_stack_root(self, _):
        self.valves_manager.maintain_stack_root(time.time(), self.stack_root_state_update_time)

    @set_ev_cls(EventFaucetEventSockHeartbeat, MAIN_DISPATCHER)
    @kill_on_exception(exc_logname)
    def _event_socket_heartbeat(self, _):
        self.valves_manager.event_socket_heartbeat()

    @set_ev_cls(EventFaucetResolveGateways, MAIN_DISPATCHER)
    @set_ev_cls(EventFaucetStateExpire, MAIN_DISPATCHER)
    @set_ev_cls(EventFaucetFastStateExpire, MAIN_DISPATCHER)
    @set_ev_cls(EventFaucetAdvertise, MAIN_DISPATCHER)
    @set_ev_cls(EventFaucetFastAdvertise, MAIN_DISPATCHER)
    @kill_on_exception(exc_logname)
    def _valve_flow_services(self, ryu_event):
        """Call a method on all Valves and send any resulting flows."""
        self.valves_manager.valve_flow_services(
            time.time(),
            self._VALVE_SERVICES[type(ryu_event)][0])

    @set_ev_cls(ofp_event.EventOFPPacketIn, MAIN_DISPATCHER)  # pylint: disable=no-member
    @kill_on_exception(exc_logname)
    def packet_in_handler(self, ryu_event):
        """Handle a packet in event from the dataplane.

        Args:
            ryu_event (ryu.controller.event.EventReplyBase): packet in message.
        """
        valve, _, msg = self._get_valve(ryu_event, require_running=True)
        if valve is None:
            return
        self.valves_manager.valve_packet_in(ryu_event.timestamp, valve, msg)

    @set_ev_cls(ofp_event.EventOFPErrorMsg, MAIN_DISPATCHER)  # pylint: disable=no-member
    @kill_on_exception(exc_logname)
    def error_handler(self, ryu_event):
        """Handle an OFPError from a datapath.

        Args:
            ryu_event (ryu.controller.ofp_event.EventOFPErrorMsg): trigger
        """
        valve, _, msg = self._get_valve(ryu_event)
        if valve is None:
            return
        valve.oferror(msg)

    @set_ev_cls(ofp_event.EventOFPSwitchFeatures, CONFIG_DISPATCHER)  # pylint: disable=no-member
    @kill_on_exception(exc_logname)
    def features_handler(self, ryu_event):
        """Handle receiving a switch features message from a datapath.

        Args:
            ryu_event (ryu.controller.ofp_event.EventOFPStateChange): trigger.
        """
        valve, ryu_dp, msg = self._get_valve(ryu_event)
        if valve is None:
            return
        self._send_flow_msgs(valve, valve.switch_features(msg), ryu_dp=ryu_dp)

    @kill_on_exception(exc_logname)
    def _datapath_connect(self, ryu_event):
        """Handle any/all re/connection of a datapath.

        Args:
            ryu_event (ryu.controller.ofp_event.Event)
        """
        now = time.time()
        valve, ryu_dp, _ = self._get_valve(ryu_event)
        if valve is None:
            return
        discovered_up_ports = {
            port.port_no for port in list(ryu_dp.ports.values())
            if (valve_of.port_status_from_state(port.state)
                and not valve_of.ignore_port(port.port_no))}
        self._send_flow_msgs(
            valve, self.valves_manager.datapath_connect(now, valve, discovered_up_ports))
        self.valves_manager.update_config_applied({valve.dp.dp_id: True})

    @kill_on_exception(exc_logname)
    def _datapath_disconnect(self, ryu_event):
        """Handle any/all disconnection of a datapath.

        Args:
            ryu_event (ryu.controller.ofp_event.Event)
        """
        valve, _, _ = self._get_valve(ryu_event)
        if valve is None:
            return
        valve.datapath_disconnect(time.time())

    @set_ev_cls(ofp_event.EventOFPDescStatsReply, MAIN_DISPATCHER)  # pylint: disable=no-member
    @kill_on_exception(exc_logname)
    def desc_stats_reply_handler(self, ryu_event):
        """Handle OFPDescStatsReply from datapath.

        Args:
            ryu_event (ryu.controller.ofp_event.EventOFPDescStatsReply): trigger.
        """
        valve, _, msg = self._get_valve(ryu_event)
        if valve is None:
            return
        valve.ofdescstats_handler(msg.body)

    @set_ev_cls(ofp_event.EventOFPPortDescStatsReply, CONFIG_DISPATCHER)  # pylint: disable=no-member
    @kill_on_exception(exc_logname)
    def port_desc_stats_reply_handler(self, ryu_event):
        """Handle OFPPortDescStatsReply from datapath.

        Args:
            ryu_event (ryu.controller.ofp_event.EventOFPPortDescStatsReply): trigger.
        """
        valve, _, msg = self._get_valve(ryu_event)
        if valve is None:
            return
        self.valves_manager.port_desc_stats_reply_handler(valve, msg, time.time())

    @set_ev_cls(ofp_event.EventOFPPortStatus, MAIN_DISPATCHER)  # pylint: disable=no-member
    @kill_on_exception(exc_logname)
    def port_status_handler(self, ryu_event):
        """Handle a port status change event.

        Args:
            ryu_event (ryu.controller.ofp_event.EventOFPPortStatus): trigger.
        """
        valve, _, msg = self._get_valve(ryu_event, require_running=True)
        if valve is None:
            return
        self.valves_manager.port_status_handler(valve, msg, time.time())

    @set_ev_cls(ofp_event.EventOFPFlowRemoved, MAIN_DISPATCHER)  # pylint: disable=no-member
    @kill_on_exception(exc_logname)
    def flowremoved_handler(self, ryu_event):
        """Handle a flow removed event.

        Args:
            ryu_event (ryu.controller.ofp_event.EventOFPFlowRemoved): trigger.
        """
        valve, ryu_dp, msg = self._get_valve(ryu_event, require_running=True)
        if valve is None:
            return
        if msg.reason == ryu_dp.ofproto.OFPRR_IDLE_TIMEOUT:
            self._send_flow_msgs(valve, valve.flow_timeout(time.time(), msg.table_id, msg.match))



#!/usr/bin/env python3

"""Manage a model of the Mininet topology for fault tolerance testing"""

import os
import networkx

from mininet.log import error


class OptimizedTopologyWatcher:
    """
    Watches multiple states of the network (both simulated and real) to calculate the
        next steps in the test to run.
    """

    # Mininet topo instance
    topo = None
    # Additional host information
    host_information = None
    # Routers configured in the test
    routers = None
    # Graph of only switch->switch links
    switch_graph = None
    # Graph of only host->switch links
    host_graph = None

    # Links that can be taken down
    eligable_links = None
    # Switches that can be taken down
    eligable_switches = None

    # List of fault names in order of occurence
    fault_list = None

    def __init__(self, topo, host_information, routers):
        """
        Args:
            topo (FaucetTopoGenerator): FaucetTopoGenerator
            host_information (dict): Additonal host information
            routers (dict): Configured routers
        """
        self.topo = topo
        self.host_information = host_information
        self.routers = routers
        self.fault_list = []

        # Create graph of only switch->switch links
        switch_graph = self.topo.g.convertTo(networkx.MultiGraph)
        for host in self.topo.hosts_by_id.values():
            switch_graph.remove_node(host)
        self.switch_graph = switch_graph

        # Create graph of only host->switch links
        host_graph = self.topo.g.convertTo(networkx.MultiGraph)
        remove_edges = []
        for edge in host_graph.edges():
            if set(edge).issubset(self.topo.switches_by_id.values()):
                remove_edges.append(edge)
        for edge in remove_edges:
            host_graph.remove_edge(*edge)
        self.host_graph = host_graph

    def add_fault(self, name):
        """
        Add a general/controller fault
        """
        error('FAULT: %s\n' % name)
        self.fault_list.append(name)

    def add_link_fault(self, src_i, dst_i, name):
        """
        Adds a link fault, i.e: removes a switch-switch edge from the predicted graph

        Args:
            src_i (int): Source index of the switch link
            dst_i (int): Destination index of the switch link
            name (str): Fault event name
        """
        try:
            self.switch_graph.remove_edge(
                self.topo.switches_by_id[src_i], self.topo.switches_by_id[dst_i])
            self.add_fault(name)
        except networkx.exception.NetworkXError:
            pass

    def add_switch_fault(self, i, name):
        """
        Add a switch fault, i.e: removes a switch node from the predicted graph

        Args:
            i (int): Index of the switch
            name (str): Fault event name
        """
        try:
            self.switch_graph.remove_node(self.topo.switches_by_id[i])
            self.add_fault(name)
        except networkx.exception.NetworkXError:
            pass

    def _get_longest_switch_path(self):
        """Return the longest simple path from comparing all possible longest simple paths"""
        # Find the longest path from all paths
        longest_path_len = 0
        longest_path = []
        for src in self.switch_graph:
            for dst in self.switch_graph:
                if src is dst:
                    continue
                for path in networkx.all_simple_paths(self.switch_graph, src, dst):
                    if len(path) > longest_path_len:
                        longest_path = path
                        longest_path_len = len(path)
                        if len(path) == len(self.switch_graph):
                            # We can stop searching as we have found the longest path
                            return longest_path
        return longest_path

    def _get_switch_connectivity_graph(self, symmetric=False, transitive=False):
        """Return connection graph of switch nodes"""
        connection_graph = networkx.MultiDiGraph()
        if not transitive:
            # Transitive assumption not present so complete all ping combinations
            for src in self.switch_graph:
                for dst in self.switch_graph:
                    connection_graph.add_edge(src, dst)
                    if not symmetric:
                        connection_graph.add_edge(dst, src)
        else:
            # Find the longest path from all paths
            longest_path = self._get_longest_switch_path()
            # Turn path into connection graph
            for i in range(0, len(longest_path) - 1):
                connection_graph.add_edge(longest_path[i], longest_path[i + 1])
                if not symmetric:
                    connection_graph.add_edge(longest_path[i + 1], longest_path[i])
            # Find and add remaining nodes
            for node in self.switch_graph:
                if node not in connection_graph:
                    # Add remaining nodes to connection graph
                    path = networkx.shortest_paths.shortest_path(
                        self.switch_graph, node, list(connection_graph.nodes())[0])
                    for i in range(0, len(path) - 1):
                        # Add path until we have reached a point that is completely inside
                        # the original simple graph
                        if path[i] in connection_graph and path[i + 1] in connection_graph:
                            break
                    connection_graph.add_edge(path[i], path[i + 1])
                    if not symmetric:
                        connection_graph.add_edge(path[i + 1], path[i])
        return connection_graph

    def get_connected_hosts(self, symmetric=False, transitive=False, intervlan_only=False):
        """
        Construct an expected connected host graph

        Args:
            symmetric (bool): Assume symmetric pings
            transitive (bool): Assume transitive pings
            intervlan_only (bool): Test hosts only inter-VLAN

        Returns:
            networkx.MultiDiGraph: expected host connection graph
        """
        # Generate switch connectivity graph
        switch_connection_graph = self._get_switch_connectivity_graph(
            symmetric=symmetric, transitive=transitive)
        # Convert switch connections to host connections
        host_connection_graph = networkx.MultiDiGraph()
        for src, dst in switch_connection_graph.edges():
            src_hosts = self.host_graph.neighbors(src)
            dst_hosts = self.host_graph.neighbors(dst)
            for src_host in src_hosts:
                for dst_host in dst_hosts:
                    if src_host is dst_host:
                        continue
                    if intervlan_only:
                        if self._routed_vlans(src_host, dst_host):
                            host_connection_graph.add_edge(src_host, dst_host)
                    else:
                        host_connection_graph.add_edge(src_host, dst_host)
        return host_connection_graph

    def _routed_vlans(self, src_host, dst_host):
        """Return true only if src_host, dst_host vlans share a router"""
        src_vlan = self.host_information[self.topo.nodeInfo(src_host)['host_n']]['vlan']
        dst_vlan = self.host_information[self.topo.nodeInfo(dst_host)['host_n']]['vlan']
        for vlans in self.routers.values():
            if src_vlan in vlans and dst_vlan in vlans:
                return True
        return False

    def get_eligable_link_events(self):
        """Return list of available stack links to take down"""
        # Remove bridges (prevents disjoint graphs)
        eligable_links = list(self.switch_graph.edges())
        graph = networkx.Graph(self.switch_graph)
        bridges = list(networkx.algorithms.bridges(graph))
        for bridge in bridges:
            if self.switch_graph.number_of_edges(*bridge) > 1:
                # Actually multiple links so not a bridge
                continue
            if bridge in eligable_links:
                eligable_links.remove(bridge)
        return eligable_links

    def get_eligable_switch_events(self):
        """Returns list of available switch names to take down"""
        # Remove bridges (prevents disjoint graphs)
        eligable_switches = list(self.switch_graph.nodes())
        graph = networkx.Graph(self.switch_graph)
        bridges = list(networkx.algorithms.bridges(graph))
        for bridge in bridges:
            if self.switch_graph.number_of_edges(*bridge) > 1:
                # Actually multiple links so not a bridge
                continue
            for node in bridge:
                if node in eligable_switches:
                    eligable_switches.remove(node)
        return eligable_switches

    def continue_faults(self):
        """Returns true whether there are more faults to occur"""
        return self.get_eligable_link_events() or self.get_eligable_switch_events()

    def dump_info(self, tmpdir):
        """Dump topology watcher info into test directory"""
        sw_graph_fn = os.path.join(tmpdir, 'final_switch_graph.txt')
        networkx.write_edgelist(self.switch_graph, sw_graph_fn)
        fault_list_fn = os.path.join(tmpdir, 'fault-list.txt')
        with open(fault_list_fn, 'w', encoding='utf-8') as fl_file:
            for fault_name in self.fault_list:
                fl_file.write(fault_name + '\n')



#!/usr/bin/env python3

"""Mininet tests for FAUCET fault-tolerance results.

 * must be run as root
 * you can run a specific test case only, by adding the class name of the test
   case to the command. Eg ./mininet_main.py FaucetUntaggedIPv4RouteTest

It is strongly recommended to run these tests via Docker, to ensure you have
all dependencies correctly installed. See ../docs/.
"""

import os
import networkx
from networkx.generators.atlas import graph_atlas_g

from clib.clib_mininet_test_main import test_main

import fault_tolerance_tests


def test_generator(func_graph, stack_roots):
    """Return the function that will start the fault-tolerance testing for a graph"""
    def test(self):
        """Test fault-tolerance of the topology"""
        self.set_up(func_graph, stack_roots)
        self.network_function()
    return test


if __name__ == '__main__':
    GRAPHS = {}
    GRAPH_ATLAS = graph_atlas_g()
    for graph in GRAPH_ATLAS:
        if (not graph
                or graph.number_of_nodes() > fault_tolerance_tests.MAX_NODES
                or graph.number_of_nodes() < fault_tolerance_tests.MIN_NODES):
            continue
        if networkx.is_connected(graph):
            GRAPHS.setdefault(graph.number_of_nodes(), [])
            GRAPHS[graph.number_of_nodes()].append(graph)
    TEST_LIMIT = int(os.getenv('FAUCET_GENERATIVE_LIMIT', '0'))

    def create_tests():
        """Create multi DP test variations."""
        test_count = 0
        for test_class in fault_tolerance_tests.TEST_CLASS_LIST:
            for test_graph in GRAPHS[test_class.NUM_DPS]:
                test_name = 'test_%s' % test_graph.name
                test_func = test_generator(test_graph, test_class.STACK_ROOTS)
                setattr(test_class, test_name, test_func)
                test_count += 1
                if TEST_LIMIT and test_count == TEST_LIMIT:
                    print('Limiting number of tests to', test_count)
                    return

    # TODO: because we are creating tests dynamically re-using the same base class,
    # we cannot run tests in parallel as tests with the same base class will
    # overwrite each other's metadata (including switch names).
    create_tests()
    test_main([fault_tolerance_tests.__name__], serial_override=True)



#!/usr/bin/env python3

"""Unit tests run as PYTHONPATH=../../.. python3 ./test_valve_dot1x.py."""

# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from faucet import faucet_dot1x

from clib.valve_test_lib import DOT1X_CONFIG, DOT1X_ACL_CONFIG, ValveTestBases


class ValveDot1xSmokeTestCase(ValveTestBases.ValveTestNetwork):
    """Smoke test to check dot1x can be initialized."""

    CONFIG = f"""
dps:
    s1:
{DOT1X_CONFIG}
        interfaces:
            p1:
                number: 1
                native_vlan: v100
                dot1x: true
            p2:
                number: 2
                output_only: True
vlans:
    v100:
        vid: 0x100
    student:
        vid: 0x200
        dot1x_assigned: True

"""

    def setUp(self):
        """Setup basic 802.1x config"""
        self.setup_valves(self.CONFIG)

    def test_get_mac_str(self):
        """Test NFV port formatter."""
        self.assertEqual('00:00:00:0f:01:01', faucet_dot1x.get_mac_str(15, 257))

    def test_handlers(self):
        """Test dot1x logoff/failure handlers."""
        valve_index = self.dot1x.dp_id_to_valve_index[self.DP_ID]
        port_no = 1
        vlan_name = 'student'
        filter_id = 'block_http'
        for handler in (self.dot1x.logoff_handler, self.dot1x.failure_handler):
            handler('0e:00:00:00:00:ff', faucet_dot1x.get_mac_str(valve_index, port_no))
        self.dot1x.auth_handler(
            '0e:00:00:00:00:ff', faucet_dot1x.get_mac_str(valve_index, port_no),
            vlan_name=vlan_name, filter_id=filter_id)


class ValveDot1xACLSmokeTestCase(ValveDot1xSmokeTestCase):
    """Smoke test to check dot1x can be initialized with dot1x ACLs."""

    ACL_CONFIG = """
acls:
    auth_acl:
        - rule:
            actions:
                allow: 1
    noauth_acl:
        - rule:
            actions:
                allow: 0
"""

    CONFIG = f"""
{ACL_CONFIG}
dps:
    s1:
{DOT1X_ACL_CONFIG}
        interfaces:
            p1:
                number: 1
                native_vlan: v100
                dot1x: true
                dot1x_acl: True
            p2:
                number: 2
                output_only: True
vlans:
    v100:
        vid: 0x100
    student:
        vid: 0x200
        dot1x_assigned: True
"""


class ValveDot1xMABSmokeTestCase(ValveDot1xSmokeTestCase):
    """Smoke test to check dot1x can be initialized with dot1x MAB."""

    CONFIG = f"""
dps:
    s1:
{DOT1X_CONFIG}
        interfaces:
            p1:
                number: 1
                native_vlan: v100
                dot1x: true
                dot1x_mab: True
            p2:
                number: 2
                output_only: True
vlans:
    v100:
        vid: 0x100
"""


class ValveDot1xDynACLSmokeTestCase(ValveDot1xSmokeTestCase):
    """Smoke test to check dot1x can be initialized with dynamic dot1x ACLs."""
    CONFIG = f"""
acls:
    accept_acl:
        dot1x_assigned: True
        rules:
        - rule:
            dl_type: 0x800      # Allow ICMP / IPv4
            ip_proto: 1
            actions:
                allow: True
        - rule:
            dl_type: 0x0806     # ARP Packets
            actions:
                allow: True
dps:
    s1:
{DOT1X_CONFIG}
        interfaces:
            p1:
                number: 1
                native_vlan: v100
                dot1x: true
                dot1x_dyn_acl: True

            p2:
                number: 2
                output_only: True
vlans:
    v100:
        vid: 0x100
"""

    def setUp(self):
        self.setup_valves(self.CONFIG)

    def test_handlers(self):
        valve_index = self.dot1x.dp_id_to_valve_index[self.DP_ID]
        port_no = 1
        vlan_name = None
        filter_id = 'accept_acl'
        for handler in (
                self.dot1x.logoff_handler,
                self.dot1x.failure_handler):
            handler(
                '0e:00:00:00:00:ff', faucet_dot1x.get_mac_str(valve_index, port_no))
        self.dot1x.auth_handler(
            '0e:00:00:00:00:ff', faucet_dot1x.get_mac_str(valve_index, port_no),
            vlan_name=vlan_name, filter_id=filter_id)


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr



"""Configuration for a datapath."""

# pylint: disable=protected-access
# pylint: disable=too-many-lines

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from collections.abc import Iterable
import copy
import random
import math
import netaddr

from faucet import faucet_pipeline
from faucet import valve_of
from faucet import valve_packet
from faucet.acl import PORT_ACL_8021X, MAB_ACL_8021X
from faucet.vlan import VLAN
from faucet.conf import Conf, test_config_condition
from faucet.faucet_pipeline import ValveTableConfig
from faucet.valve import SUPPORTED_HARDWARE
from faucet.valve_table import ValveTable, ValveGroupTable
from faucet.stack import Stack


# Documentation generated using documentation_generator.py
# For attributues to be included in documentation they must
# have a default value, and their descriptor must come
# immediately after being set. See below for example.
class DP(Conf):
    """Stores state related to a datapath controlled by Faucet, including
configuration.
"""
    DEFAULT_LLDP_SEND_INTERVAL = 5
    DEFAULT_LLDP_MAX_PER_INTERVAL = 5
    mutable_attrs = frozenset(['vlans'])

    # Values that are set to None will be set using set_defaults
    # they are included here for testing and informational purposes
    defaults = {
        'dp_id': None,
        # Name for this dp, used for stats reporting and configuration
        'name': None,
        'interfaces': {},
        'interface_ranges': {},
        # How much to offset default priority by
        'priority_offset': 0,
        # Some priority values
        'lowest_priority': None,
        'low_priority': None,
        'high_priority': None,
        'highest_priority': None,
        'cookie': 1524372928,
        # Identification cookie value to allow for multiple controllers to control the same datapath
        'timeout': 300,
        # inactive MAC timeout
        'description': None,
        # description, strictly informational
        'hardware': 'Open vSwitch',
        # The hardware maker (for chosing an openflow driver)
        'arp_neighbor_timeout': 30,
        # ARP neighbor timeout (seconds)
        'nd_neighbor_timeout': 30,
        # IPv6 ND neighbor timeout (seconds)
        'ofchannel_log': None,
        # OF channel log
        'stack': None,
        # stacking config, when cross connecting multiple DPs
        'ignore_learn_ins': 10,
        # Ignore every approx nth packet for learning.
        # 2 will ignore 1 out of 2 packets; 3 will ignore 1 out of 3 packets.
        # This limits control plane activity when learning new hosts rapidly.
        # Flooding will still be done by the dataplane even with a packet
        # is ignored for learning purposes.
        'drop_broadcast_source_address': True,
        # By default drop packets with a broadcast source address
        'drop_spoofed_faucet_mac': True,
        # By default drop packets on datapath spoofing the FAUCET_MAC
        'group_table': False,
        # Use GROUP tables for VLAN flooding
        'max_hosts_per_resolve_cycle': 5,
        # Max hosts to try to resolve per gateway resolution cycle.
        'max_host_fib_retry_count': 10,
        # Max number of times to retry resolution of a host FIB route.
        'max_resolve_backoff_time': 64,
        # Max number of seconds to back off to when resolving nexthops.
        'packetin_pps': None,
        # Ask switch to rate limit packetin pps. TODO: Not supported by OVS in 2.7.0
        'slowpath_pps': None,
        # Ask switch to rate limit slowpath pps. TODO: Not supported by OVS in 2.7.0
        'learn_jitter': 0,
        # Jitter learn timeouts by up to this many seconds
        'learn_ban_timeout': 0,
        # When banning/limiting learning, wait this many seconds before learning can be retried
        'advertise_interval': 30,
        # How often to slow advertise (eg. IPv6 RAs)
        'fast_advertise_interval': 5,
        # How often to fast advertise (eg. LACP)
        'proactive_learn_v4': True,
        # whether proactive learning is enabled for IPv4 nexthops
        'proactive_learn_v6': True,
        # whether proactive learning is enabled for IPv6 nexthops
        'use_idle_timeout': False,
        # Turn on/off the use of idle timeout for src_table, default OFF.
        'lldp_beacon': {},
        # Config for LLDP beacon service.
        'metrics_rate_limit_sec': 0,
        # Rate limit metric updates if last update was less than this many seconds ago.
        'faucet_dp_mac': valve_packet.FAUCET_MAC,
        # MAC address of packets sent by FAUCET, not associated with any VLAN.
        'combinatorial_port_flood': False,
        # if True, use a seperate output flow for each input port on this VLAN.
        'lacp_timeout': 30,
        # Number of seconds without a LACP message when we consider a LACP group down.
        'dp_acls': None,
        # List of dataplane ACLs (overriding per port ACLs).
        'dot1x': {},
        # Experimental dot1x configuration.
        'table_sizes': {},
        # Table sizes for TFM switches.
        'min_wildcard_table_size': 32,
        # Minimum table size for wildcard tables.
        'max_wildcard_table_size': 1024 + 256,
        # Maximum table size for wildcard tables.
        'port_table_scale_factor': 1.0,
        # Amount to scale port scaled table sizes by.
        'global_vlan': 0,
        # Reserved VID for internal global router VLAN.
        'cache_update_guard_time': 0,
        # Don't update L2 cache if port didn't change within this many seconds (default timeout/2).
        'use_classification': False,
        # Don't update L2 cache if port didn't change within this many seconds.
        'egress_pipeline': False,
        # Experimental inclusion of an egress pipeline
        'strict_packet_in_cookie': True,
        # Apply strict packet in checking to all packet ins.
        'multi_out': True,
        # Have OFA copy packet outs to multiple ports.
        'idle_dst': True,
        # If False, workaround for flow idle timer not reset on flow refresh.
    }

    defaults_types = {
        'dp_id': int,
        'name': str,
        'interfaces': dict,
        'interface_ranges': dict,
        'priority_offset': int,
        'lowest_priority': int,
        'low_priority': int,
        'high_priority': int,
        'highest_priority': int,
        'cookie': int,
        'timeout': int,
        'description': str,
        'hardware': str,
        'arp_neighbor_timeout': int,
        'nd_neighbor_timeout': int,
        'ofchannel_log': str,
        'stack': dict,
        'ignore_learn_ins': int,
        'drop_broadcast_source_address': bool,
        'drop_spoofed_faucet_mac': bool,
        'group_table': bool,
        'max_hosts_per_resolve_cycle': int,
        'max_host_fib_retry_count': int,
        'max_resolve_backoff_time': int,
        'packetin_pps': int,
        'slowpath_pps': int,
        'learn_jitter': int,
        'learn_ban_timeout': int,
        'advertise_interval': int,
        'fast_advertise_interval': int,
        'proactive_learn_v4': bool,
        'proactive_learn_v6': bool,
        'use_idle_timeout': bool,
        'lldp_beacon': dict,
        'metrics_rate_limit_sec': int,
        'faucet_dp_mac': str,
        'combinatorial_port_flood': bool,
        'dp_acls': list,
        'dot1x': dict,
        'table_sizes': dict,
        'min_wildcard_table_size': int,
        'max_wildcard_table_size': int,
        'port_table_scale_factor': float,
        'global_vlan': int,
        'cache_update_guard_time': int,
        'use_classification': bool,
        'egress_pipeline': bool,
        'strict_packet_in_cookie': bool,
        'multi_out': bool,
        'lacp_timeout': int,
        'idle_dst': bool,
    }

    default_table_sizes_types = {
        'port_acl': int,
        'vlan': int,
        'vlan_acl': int,
        'classification': int,
        'eth_src': int,
        'ipv4_fib': int,
        'ipv6_fib': int,
        'vip': int,
        'eth_dst_hairpin': int,
        'eth_dst': int,
        'flood': int,
    }

    lldp_beacon_defaults_types = {
        'send_interval': int,
        'max_per_interval': int,
        'system_name': str,
    }

    dot1x_defaults_types = {
        'nfv_intf': str,
        'nfv_sw_port': int,
        'radius_ip': str,
        'radius_port': int,
        'radius_secret': str,
        'auth_acl': str,
        'noauth_acl': str,
    }

    def __init__(self, _id, dp_id, conf):
        """Constructs a new DP object"""
        self.acls = None
        self.acls_in = None
        self.advertise_interval = None
        self.fast_advertise_interval = None
        self.arp_neighbor_timeout = None
        self.nd_neighbor_timeout = None
        self.combinatorial_port_flood = None
        self.configured = False
        self.cookie = None
        self.description = None
        self.dot1x = {}
        self.dp_acls = None
        self.dp_id = None
        self.drop_broadcast_source_address = None
        self.drop_spoofed_faucet_mac = None
        self.dyn_last_coldstart_time = None
        self.dyn_running = False
        self.dyn_up_port_nos = None
        self.egress_pipeline = None
        self.faucet_dp_mac = None
        self.global_vlan = None
        self.groups = None
        self.group_table = False
        self.hardware = None
        self.high_priority = None
        self.highest_priority = None
        self.ignore_learn_ins = None
        self.interface_ranges = None
        self.interfaces = None
        self.lacp_timeout = None
        self.learn_ban_timeout = None
        self.learn_jitter = None
        self.lldp_beacon = None
        self.low_priority = None
        self.lowest_priority = None
        self.max_host_fib_retry_count = None
        self.max_hosts_per_resolve_cycle = None
        self.max_resolve_backoff_time = None
        self.meters = None
        self.all_meters = None
        self.metrics_rate_limit_sec = None
        self.name = None
        self.ofchannel_log = None
        self.output_only_ports = None
        self.packetin_pps = None
        self.slowpath_pps = None
        self.ports = None
        self.priority_offset = None
        self.proactive_learn_v4 = None
        self.proactive_learn_v6 = None
        self.proactive_nd_limit = None
        self.routers = None
        self.stack = None
        self.tables = None
        self.timeout = None
        self.unicast_flood = None
        self.use_idle_timeout = None
        self.vlans = None
        self.min_wildcard_table_size = None
        self.max_wildcard_table_size = None
        self.port_table_scale_factor = None
        self.cache_update_guard_time = None
        self.use_classification = None
        self.strict_packet_in_cookie = None
        self.multi_out = None
        self.idle_dst = None
        self.has_acls = None

        self.acls = {}
        self.vlans = {}
        self.ports = {}
        self.routers = {}
        self.hairpin_ports = []
        self.output_only_ports = []
        self.lldp_beacon_ports = []
        self.lacp_active_ports = []
        self.tables = {}
        self.meters = {}
        self.all_meters = {}
        self.lldp_beacon = {}
        self.table_sizes = {}
        self.dyn_up_port_nos = set()
        self.has_externals = None
        self.tunnel_acls = []

        super().__init__(_id, dp_id, conf)

    def __str__(self):
        return str(self.name)

    def clone_dyn_state(self, prev_dp, dps=None):
        """Clone dynamic state for this dp"""
        self.dyn_running = prev_dp.dyn_running
        self.dyn_up_port_nos = set(prev_dp.dyn_up_port_nos)
        self.dyn_last_coldstart_time = prev_dp.dyn_last_coldstart_time
        for number in self.ports:
            self.ports[number].clone_dyn_state(prev_dp.ports.get(number))
        if self.stack:
            self.stack.clone_dyn_state(prev_dp.stack, dps)

    def cold_start(self, now):
        """Update to reflect a cold start"""
        self.dyn_last_coldstart_time = now
        self.dyn_running = True
        for vlan in self.vlans.values():
            vlan.reset_caches()

    def check_config(self):
        """Check configuration of this dp"""
        super().check_config()
        test_config_condition(not isinstance(self.dp_id, int), (
            f'dp_id must be {int} not {type(self.dp_id)}'))
        test_config_condition(self.dp_id < 0 or self.dp_id > 2**64 - 1, (
            f'DP ID {self.dp_id} not in valid range'))
        test_config_condition(not netaddr.valid_mac(self.faucet_dp_mac), (
            f'invalid MAC address {self.faucet_dp_mac}'))
        test_config_condition(not (self.interfaces or self.interface_ranges), (
            f'DP {self} must have at least one interface'))
        test_config_condition(self.timeout < 15, 'timeout must be > 15')
        test_config_condition(self.timeout > 65535, 'timeout cannot be > than 65335')
        # To prevent L2 learning from timing out before L3 can refresh
        test_config_condition(not (self.arp_neighbor_timeout < (self.timeout / 2)), (
            'L2 timeout must be > ARP timeout * 2'))
        test_config_condition(
            self.arp_neighbor_timeout > 65535, 'arp_neighbor_timeout cannot be > 65535')
        test_config_condition(not (self.nd_neighbor_timeout < (self.timeout / 2)), (
            'L2 timeout must be > ND timeout * 2'))
        test_config_condition(
            self.nd_neighbor_timeout > 65535, 'nd_neighbor_timeout cannot be > 65535')
        test_config_condition(self.combinatorial_port_flood and self.group_table, (
            'combinatorial_port_flood and group_table mutually exclusive'))
        if self.cache_update_guard_time == 0:
            self.cache_update_guard_time = int(self.timeout / 2)
        if self.learn_jitter == 0:
            self.learn_jitter = int(max(math.sqrt(self.timeout) * 3, 1))
        if self.learn_ban_timeout == 0:
            self.learn_ban_timeout = self.learn_jitter
        if self.lldp_beacon:
            self._lldp_defaults()
        if self.dot1x:
            self._check_conf_types(self.dot1x, self.dot1x_defaults_types)
        self._check_conf_types(self.table_sizes, self.default_table_sizes_types)
        self.stack = Stack('stack', self.dp_id, self.name,
                           self.canonical_port_order, self.lacp_down_ports, self.lacp_ports,
                           self.stack)

    def _lldp_defaults(self):
        self._check_conf_types(self.lldp_beacon, self.lldp_beacon_defaults_types)
        if 'send_interval' not in self.lldp_beacon:
            self.lldp_beacon['send_interval'] = self.DEFAULT_LLDP_SEND_INTERVAL
        test_config_condition(self.lldp_beacon['send_interval'] < 1, (
            f'DP ID {self.dp_id} LLDP beacon send_interval not in valid range'))
        if 'max_per_interval' not in self.lldp_beacon:
            self.lldp_beacon['max_per_interval'] = self.DEFAULT_LLDP_MAX_PER_INTERVAL
        self.lldp_beacon = self._set_unknown_conf(
            self.lldp_beacon, self.lldp_beacon_defaults_types)
        if self.lldp_beacon.get('system_name', None) is None:
            self.lldp_beacon['system_name'] = self.name

    def _generate_acl_tables(self):
        all_acls = {}
        if self.dot1x:
            # NOTE: All acl's are added to the acl list and then referred to later by ports
            acls = [PORT_ACL_8021X, MAB_ACL_8021X,
                    self.acls.get(self.dot1x.get('auth_acl'), None),
                    self.acls.get(self.dot1x.get('noauth_acl'), None)]

            acls.extend([acl for acl_name, acl in self.acls.items() if acl.dot1x_assigned])
            all_acls['port_acl'] = [acl for acl in acls if acl is not None]

        for vlan in self.vlans.values():
            if vlan.acls_in:
                all_acls.setdefault('vlan_acl', [])
                all_acls['vlan_acl'].extend(vlan.acls_in)
            if vlan.acls_out:
                all_acls.setdefault('egress_acl', [])
                all_acls['egress_acl'].extend(vlan.acls_out)
                self.egress_pipeline = True
        if self.dp_acls:
            test_config_condition(self.dot1x, (
                'DP ACLs and 802.1x cannot be configured together'))
            all_acls.setdefault('port_acl', [])
            all_acls['port_acl'].extend(self.dp_acls)
        else:
            for port in self.ports.values():
                if port.acls_in:
                    test_config_condition(port.dot1x, (
                        'port ACLs and 802.1x cannot be configured together'))
                    all_acls.setdefault('port_acl', [])
                    all_acls['port_acl'].extend(port.acls_in)
                if self.dot1x and port.number == self.dot1x['nfv_sw_port']:
                    test_config_condition(not port.output_only, (
                        'NFV Ports must have output_only set to True.'
                    ))
        if self.tunnel_acls:
            all_acls.setdefault('port_acl', [])
            all_acls['port_acl'].extend(self.tunnel_acls)
        table_config = {}
        for table_name, acls in all_acls.items():
            matches = {}
            set_fields = set()
            meter = False
            exact_match = False
            default = faucet_pipeline.DEFAULT_CONFIGS[table_name]
            for acl in acls:
                for field, has_mask in acl.matches.items():
                    if has_mask or field not in matches:
                        matches[field] = has_mask
                set_fields.update(acl.set_fields)
                meter = meter or acl.meter
                exact_match = acl.exact_match
            table_config[table_name] = ValveTableConfig(
                table_name,
                default.table_id,
                exact_match=exact_match,
                meter=meter,
                output=True,
                match_types=tuple(sorted(matches.items())),
                set_fields=tuple(sorted(set_fields)),
                next_tables=default.next_tables)
        # TODO: dynamically configure output attribute
        return table_config

    def pipeline_str(self):
        """Text description of pipeline."""
        table_configs = sorted([
            (table.table_id, str(table.table_config))
            for table in self.tables.values()])
        return '\n'.join([
            f'table ID {table_id} {table_config}'
            for table_id, table_config in table_configs])

    def pipeline_tableids(self):
        """Return pipeline table IDs."""
        return {table.table_id for table in self.tables.values()}

    def _configure_tables(self):
        """Configure FAUCET pipeline with tables."""
        valve_cl = SUPPORTED_HARDWARE.get(self.hardware, None)
        test_config_condition(
            not valve_cl, f'hardware {self.hardware} must be in {list(SUPPORTED_HARDWARE)}')
        if valve_cl is None:
            return

        tables = {}
        self.groups = ValveGroupTable()
        relative_table_id = 0
        included_tables = copy.deepcopy(faucet_pipeline.MINIMUM_FAUCET_PIPELINE_TABLES)
        acl_tables = self._generate_acl_tables()
        if acl_tables:
            included_tables.update(set(acl_tables.keys()))
            self.has_acls = True
        # Only configure IP routing tables if enabled.
        for vlan in self.vlans.values():
            for ipv in vlan.ipvs():
                included_tables.add(f'ipv{ipv}_fib')
                included_tables.add('vip')
        if valve_cl.STATIC_TABLE_IDS:
            included_tables.add('port_acl')
            self.has_acls = True
        if self.hairpin_ports:
            included_tables.add('eth_dst_hairpin')
        if self.use_classification:
            included_tables.add('classification')
        if self.egress_pipeline:
            included_tables.add('egress')
        if self.coprocessor_ports():
            included_tables.add('copro')
        canonical_configs = [
            config for config in faucet_pipeline.FAUCET_PIPELINE
            if config.name in included_tables]
        table_configs = {}
        for relative_table_id, canonical_table_config in enumerate(canonical_configs, start=0):
            name = canonical_table_config.name
            table_config = acl_tables.get(
                name, copy.deepcopy(canonical_table_config))
            if not self.egress_pipeline:
                table_config.metadata_write = 0
                table_config.metadata_match = 0
            if not valve_cl.STATIC_TABLE_IDS:
                table_config.table_id = relative_table_id
            table_configs[name] = table_config

        # Stacking with external ports, so need external forwarding request field.
        if self.has_externals:
            for table_name in ('vlan', 'eth_dst', 'flood'):
                table = table_configs[table_name]
                table.match_types += ((valve_of.EXTERNAL_FORWARDING_FIELD, False),)
                if table.set_fields is not None:
                    table.set_fields += (valve_of.EXTERNAL_FORWARDING_FIELD,)
                else:
                    table.set_fields = (valve_of.EXTERNAL_FORWARDING_FIELD,)

        if self.restricted_bcast_arpnd_ports():
            table_configs['flood'].match_types += (('eth_type', False),)

        if 'egress_acl' in included_tables:
            table_configs['eth_dst'].miss_goto = 'egress_acl'

        oxm_fields = set(valve_of.MATCH_FIELDS.keys())

        for table_name, table_config in table_configs.items():
            if table_config.set_fields:
                set_fields = set(table_config.set_fields)
                test_config_condition(
                    not set_fields.issubset(oxm_fields),
                    f'set_fields not all OpenFlow OXM fields {set_fields - oxm_fields}')
            if table_config.match_types:
                matches = set(match for match, _ in table_config.match_types)
                test_config_condition(
                    not matches.issubset(oxm_fields),
                    f'matches not all OpenFlow OXM fields {matches - oxm_fields}')

            scale_factor = 1.0
            # Need flows for internal/external.
            if self.has_externals:
                scale_factor *= 2

            # Table scales with number of VLANs only.
            if table_config.vlan_scale:
                scale_factor *= (len(self.vlans) * table_config.vlan_scale)

            # Table scales with number of ports and VLANs.
            elif table_config.vlan_port_scale:
                scale_factor *= (len(self.vlans) * len(self.ports) * table_config.vlan_port_scale)
                scale_factor *= self.port_table_scale_factor

                if table_config.name == 'flood':
                    # We need flows for all ports when using combinatorial port flood.
                    if self.combinatorial_port_flood:
                        scale_factor *= len(self.ports)
                    # We need more flows for more broadcast rules.
                    if self.restricted_bcast_arpnd_ports():
                        scale_factor *= 2

            # Always multiple of min_wildcard_table_size
            table_size_multiple = int(scale_factor / self.min_wildcard_table_size) + 1
            size = table_size_multiple * self.min_wildcard_table_size

            if not table_config.exact_match:
                size = max(size, self.min_wildcard_table_size)
                size = min(size, self.max_wildcard_table_size)

            # Hard override for size if present.
            size = self.table_sizes.get(table_name, size)

            table_config.size = size
            table_config.next_tables = [
                tbl_name for tbl_name in table_config.next_tables
                if tbl_name in table_configs]
            next_table_ids = [
                table_configs[tbl_name].table_id for tbl_name in table_config.next_tables]
            tables[table_name] = ValveTable(
                table_name, table_config, self.cookie,
                notify_flow_removed=self.use_idle_timeout,
                next_tables=next_table_ids
            )
        self.tables = tables

    def set_defaults(self):
        super().set_defaults()
        self._set_default('dp_id', self._id)
        self._set_default('name', str(self._id))
        self._set_default('lowest_priority', self.priority_offset)
        self._set_default('low_priority', self.priority_offset + 9000)
        self._set_default('high_priority', self.low_priority + 1)
        self._set_default('highest_priority', self.high_priority + 98)
        self._set_default('description', self.name)

    def table_by_id(self, table_id):
        """Gets first table with table id"""
        tables = [table for table in self.tables.values() if table_id == table.table_id]
        if tables:
            return tables[0]
        return None

    def port_no_valid(self, port_no):
        """Return True if supplied port number valid on this datapath."""
        return not valve_of.ignore_port(port_no) and port_no in self.ports

    def base_prom_labels(self):
        """Return base Prometheus labels for this DP."""
        return dict(dp_id=hex(self.dp_id), dp_name=self.name)

    def port_labels(self, port_no):
        """Return port name and description labels for a port number."""
        port_name = str(port_no)
        port_description = None
        if port_no in self.ports:
            port = self.ports[port_no]
            port_name = port.name
            port_description = port.description
        elif port_no == valve_of.ofp.OFPP_CONTROLLER:
            port_name = 'CONTROLLER'
        elif port_no == valve_of.ofp.OFPP_LOCAL:
            port_name = 'LOCAL'
        if port_description is None:
            port_description = port_name
        return dict(self.base_prom_labels(), port=port_name, port_description=port_description)

    def classification_table(self):
        """Returns classification table"""
        if self.use_classification:
            return self.tables['classification']
        return self.tables['eth_src']

    def output_tables(self):
        """Return tables that cause a packet to be forwarded."""
        if self.hairpin_ports:
            return (self.tables['eth_dst_hairpin'], self.tables['eth_dst'])
        return (self.tables['eth_dst'],)

    def output_table(self):
        """Returns first output table"""
        return self.output_tables()[0]

    def match_tables(self, match_type):
        """Return list of tables with matches of a specific match type."""
        return [
            table for table in self.tables.values()
            if table.match_types is None or match_type in table.match_types]

    def non_vlan_ports(self):
        """Ports that don't have VLANs on them."""
        ports = set()
        for non_vlan in (self.output_only_ports, self.stack_ports(), self.coprocessor_ports()):
            ports.update(set(non_vlan))
        return ports

    def stack_ports(self):
        """Return list of stack ports"""
        if self.stack:
            return tuple(self.stack.ports)
        return []

    def coprocessor_ports(self):
        """Return list of coprocessor ports."""
        return tuple(port for port in self.ports.values() if port.coprocessor)

    def restricted_bcast_arpnd_ports(self):
        """Return ports that have restricted broadcast set."""
        return tuple(port for port in self.ports.values() if port.restricted_bcast_arpnd)

    def lacp_ports(self):
        """Return ports that have LACP."""
        return tuple(port for port in self.ports.values() if port.lacp)

    def lacp_up_ports(self):
        """Return ports that have LACP up."""
        return tuple(port for port in self.lacp_ports() if port.is_actor_up())

    def lacp_down_ports(self):
        """Return ports that have LACP not UP"""
        return tuple(port for port in self.lacp_ports() if not port.is_actor_up())

    def lacp_nosync_ports(self):
        """Return ports that have LACP status NO_SYNC."""
        return tuple(port for port in self.lacp_ports() if port.is_actor_nosync())

    def lags(self):
        """Return dict of LAGs mapped to member ports."""
        lags = defaultdict(list)
        for port in self.lacp_ports():
            lags[port.lacp].append(port)
        return lags

    def lags_up(self):
        """Return dict of LAGs mapped to member ports that have LACP up."""
        lags = defaultdict(list)
        for port in self.lacp_up_ports():
            lags[port.lacp].append(port)
        return lags

    def lags_nosync(self):
        """Return dict of LAGs mapped to member ports that have LACP in NO SYNC."""
        lags = defaultdict(list)
        for port in self.lacp_nosync_ports():
            lags[port.lacp].append(port)
        return lags

    def all_lags_up(self):
        """Return True if all LAGs have at least one port up."""
        return set(self.lags()) == set(self.lags_up())

    def add_acl(self, acl_ident, acl):
        """Add an ACL to this DP."""
        self.acls[acl_ident] = acl

    def add_router(self, router_ident, router):
        """Add a router to this DP."""
        self.routers[router_ident] = router

    def add_port(self, port):
        """Add a port to this DP."""
        port_num = port.number
        self.ports[port_num] = port
        if port.output_only:
            self.output_only_ports.append(port)
        if port.stack:
            self.stack.add_port(port)
        if port.lldp_beacon_enabled():
            self.lldp_beacon_ports.append(port)
        if port.hairpin or port.hairpin_unicast:
            self.hairpin_ports.append(port)
        if port.lacp and port.lacp_active:
            self.lacp_active_ports.append(port)

    def lldp_beacon_send_ports(self, now):
        """Return list of ports to send LLDP packets; stacked ports always send LLDP."""
        send_ports = []
        if self.lldp_beacon:
            priority_ports = {
                port for port in self.stack_ports()
                if port.running() and port.lldp_beacon_enabled()}
            cutoff_beacon_time = now - self.lldp_beacon['send_interval']
            nonpriority_ports = {
                port for port in self.lldp_beacon_ports
                if port.running() and (
                    port.dyn_last_lldp_beacon_time is None
                    or port.dyn_last_lldp_beacon_time < cutoff_beacon_time)}
            nonpriority_ports -= priority_ports
            send_ports.extend(list(priority_ports))
            nonpriority_ports = list(nonpriority_ports)
            random.shuffle(nonpriority_ports)
            nonpriority_ports = nonpriority_ports[:self.lldp_beacon['max_per_interval']]
            send_ports.extend(nonpriority_ports)
        return send_ports

    def resolve_stack_topology(self, dps, meta_dp_state):
        """Resolve inter-DP config for stacking"""
        if self.stack:
            self.stack.resolve_topology(dps, meta_dp_state)
            for dp in dps:
                # Must set externals flag for entire stack.
                if dp.stack and dp.has_externals:
                    self.has_externals = True
                    break
            self.finalize_tunnel_acls(dps)

    def finalize_tunnel_acls(self, dps):
        """Resolve each tunnels sources"""
        # Find all tunnel ACLs that are in `self.tunnel_acls` that are have a configured source
        if self.tunnel_acls:
            # TODO: A Tunnel ACL can contain multiple different tunnel IDs
            tunnel_ids = {tunnel_acl._id: tunnel_acl for tunnel_acl in self.tunnel_acls}
            referenced_acls = set()
            for dp in dps:
                if dp.dp_acls:
                    for acl in dp.dp_acls:
                        tunnel_acl = tunnel_ids.get(acl._id)
                        if tunnel_acl:
                            # ACL is configured on a DP
                            tunnel_acl.add_tunnel_source(dp.name, None)
                            referenced_acls.add(tunnel_acl._id)
                else:
                    for port in dp.ports.values():
                        if port.acls_in:
                            for acl in port.acls_in:
                                tunnel_acl = tunnel_ids.get(acl._id)
                                if tunnel_acl:
                                    tunnel_acl.add_tunnel_source(dp.name, port.number)
                                    referenced_acls.add(tunnel_acl._id)
            # Any tunnel ACL that has not been resolved should be ignored
            for tunnel_id, tunnel_acl in tunnel_ids.items():
                if tunnel_id not in referenced_acls:
                    self.tunnel_acls.remove(tunnel_acl)

    @staticmethod
    def canonical_port_order(ports):
        """Return iterable of ports in consistent order."""
        return sorted(ports, key=lambda x: x.number)

    def reset_refs(self, vlans=None):
        """Resets VLAN references."""

        if vlans is None:
            vlans = self.vlans
            router_vlans = {vlan._id for router in self.routers.values() for vlan in router.vlans}
        else:
            router_vlans = {vlan for router in self.routers.values() for vlan in router.vlans}

        vlan_ports = defaultdict(set)
        for port in self.ports.values():
            for vlan in port.vlans():
                vlan_ports[vlan].add(port)

        if self.stack_ports or self.stack.is_root():
            new_vlans = list(vlans.values())
        else:
            new_vlans = []
            for vlan in vlans.values():
                if (vlan_ports[vlan] or vlan.reserved_internal_vlan
                        or vlan.dot1x_assigned or vlan._id in router_vlans):
                    new_vlans.append(vlan)

        self.vlans = {}
        for vlan in new_vlans:
            vlan.reset_ports(vlan_ports[vlan])
            self.vlans[vlan.vid] = vlan

    def resolve_port(self, port_name):
        """Resolve a port by number or name."""
        if isinstance(port_name, int):
            if port_name in self.ports:
                return self.ports[port_name]
        elif isinstance(port_name, str):
            resolved_ports = [port for port in self.ports.values() if port_name == port.name]
            if resolved_ports:
                return resolved_ports[0]
        return None

    def finalize_config(self, dps):
        """Perform consistency checks after initial config parsing."""

        dp_by_name = {}
        vlan_by_name = {}
        acl_meters = set()

        def first_unused_vlan_id(vid):
            """Returns the first unused VID from the starting vid"""
            used_vids = sorted([vlan.vid for vlan in self.vlans.values()])
            while vid in used_vids:
                vid += 1
            return vid

        def create_vlan(vid):
            """Creates a VLAN object with the VID"""
            test_config_condition(vid in self.vlans, (
                'Attempting to dynamically create a VLAN with ID that already exists'))
            vlan = VLAN(vid, self.dp_id, None)
            self.vlans[vlan.vid] = vlan
            return vlan

        def resolve_ports(port_names):
            """Resolve list of ports, by port by name or number."""
            resolved_ports = []
            for port_name in port_names:
                port = self.resolve_port(port_name)
                if port is not None:
                    resolved_ports.append(port)
            return resolved_ports

        def resolve_vlan(vlan_name):
            """Resolve VLAN by name or VID."""
            test_config_condition(not isinstance(vlan_name, (str, int)), (
                f'VLAN must be type {str} or {int} not {type(vlan_name)}'))
            if vlan_name in vlan_by_name:
                return vlan_by_name[vlan_name]
            if vlan_name in self.vlans:
                return self.vlans[vlan_name]
            return None

        def resolve_vlans(vlan_names):
            """Resolve a list of VLAN names."""
            vlans = []
            for vlan_name in vlan_names:
                vlan = resolve_vlan(vlan_name)
                if vlan:
                    vlans.append(vlan)
            return vlans

        def resolve_stack_dps():
            """Resolve DP references in stacking config."""
            if self.stack_ports():
                port_stack_dp = {}
                for port in self.stack_ports():
                    stack_dp = port.stack['dp']
                    test_config_condition(stack_dp not in dp_by_name, (
                        f'stack DP {stack_dp} not defined'))
                    port_stack_dp[port] = dp_by_name[stack_dp]
                for port, dp in port_stack_dp.items():
                    port.stack['dp'] = dp
                    stack_port = dp.resolve_port(port.stack['port'])
                    test_config_condition(stack_port is None, (
                        f'stack port {port.stack["port"]} not defined in DP{dp.name}'))
                    port.stack['port'] = stack_port

        def resolve_mirror_destinations():
            """Resolve mirror port references and destinations."""
            mirror_from_port = defaultdict(list)
            for mirror_port in self.ports.values():
                if mirror_port.mirror is not None:
                    mirrored_ports = resolve_ports(mirror_port.mirror)
                    test_config_condition(len(mirrored_ports) != len(mirror_port.mirror), (
                        f'port mirror not defined in DP {self.name}'))
                    for mirrored_port in mirrored_ports:
                        mirror_from_port[mirrored_port].append(mirror_port)

            # TODO: confusingly, mirror at config time means what ports to mirror from.
            #   But internally we use as a list of ports to mirror to.
            for mirrored_port, mirror_ports in mirror_from_port.items():
                mirrored_port.mirror = []
                for mirror_port in mirror_ports:
                    mirrored_port.mirror.append(mirror_port.number)
                    if not mirror_port.coprocessor:
                        mirror_port.output_only = True

        def resolve_acl(acl_in, vid=None, port_num=None):
            """
            Resolve an individual ACL
            Args:
                acl_in (str): ACL name to find reference in the acl list
                vid (int): VID of the VLAN the ACL is being applied to
                port_num (int): The number of the port the ACl is being applied to
            Returns:
                matches, set_fields, meter (3-Tuple): ACL matches, set fields and meter values
            """
            test_config_condition(acl_in not in self.acls, (
                f'missing ACL {acl_in} in DP:{self.name}'))
            acl = self.acls[acl_in]
            tunnel_dsts_to_vlan = {}

            def resolve_port_cb(port_name):
                """Resolve port"""
                port = self.resolve_port(port_name)
                if port:
                    return port.number
                return port

            def get_tunnel_vlan(tunnel_id_name, resolved_dst):
                """
                Obtain the VLAN that is configured for a tunnel.
                If the tunnel VLAN exists, ensure it has the correct properties and is not used.
                If the VLAN does not exist, then create one.

                Args:
                    tunnel_id_name (str/int/None): Reference to VLAN object that the tunnel will use
                    resolved_dst (tuple): DP, port destination tuple
                Returns:
                    VLAN: VLAN object used by the tunnel
                """
                if not tunnel_id_name:
                    if resolved_dst in tunnel_dsts_to_vlan:
                        tunnel_vlan = tunnel_dsts_to_vlan[resolved_dst]
                    else:
                        # Create a VLAN using the first unused VLAN ID
                        # Get highest non-reserved VLAN
                        vlan_offset = max([
                            vlan.vid for vlan in self.vlans.values()
                            if not vlan.reserved_internal_vlan])
                        # Also need to account for the potential number of tunnels
                        ordered_acls = sorted(self.acls)
                        index = ordered_acls.index(acl_in) + 1
                        acl_tunnels = [self.acls[name].get_num_tunnels() for name in ordered_acls]
                        tunnel_offset = sum(acl_tunnels[:index])
                        start_pos = vlan_offset + tunnel_offset
                        tunnel_vid = first_unused_vlan_id(start_pos)
                        tunnel_vlan = create_vlan(tunnel_vid)
                        tunnel_vlan.reserved_internal_vlan = True
                else:
                    # Tunnel ID has been specified, so search for the VLAN
                    tunnel_vlan = resolve_vlan(tunnel_id_name)
                    if tunnel_vlan:
                        # VLAN exists, i.e: user specified the VLAN so check if it is reserved
                        test_config_condition(not tunnel_vlan.reserved_internal_vlan, (
                            f'VLAN {tunnel_vlan.name} is required for use by'
                            f' tunnel {tunnel_id_name} but is not reserved'))
                    else:
                        # VLAN does not exist, so the ID should be the VID the user wants
                        test_config_condition(isinstance(tunnel_id_name, str), (
                            f'Tunnel VLAN ({tunnel_id_name}) does not exist'))
                        # Create the tunnel VLAN object
                        tunnel_vlan = create_vlan(tunnel_id_name)
                        tunnel_vlan.reserved_internal_vlan = True
                    existing_tunnel_vlan = tunnel_dsts_to_vlan.get(resolved_dst, None)
                    if existing_tunnel_vlan is not None:
                        test_config_condition(
                            existing_tunnel_vlan == tunnel_vlan.vid,
                            f'Cannot have multiple tunnel IDs ({existing_tunnel_vlan.vid},'
                            f' {tunnel_vlan.vid}) to same destination {resolved_dst}')
                return tunnel_vlan

            def resolve_tunnel_objects(dst_dp_name, dst_port_name, tunnel_id_name):
                """
                Resolves the names of the tunnel src and dst (DP & port) pairs into the correct \
                    objects
                Args:
                    dst_dp (str): DP of the tunnel's destination port
                    dst_port (int/None): Destination port of the tunnel
                    tunnel_id_name (int/str/None): Tunnel identification number or VLAN reference
                Returns:
                    dst_dp name, dst_port name and tunnel id
                """
                # VLAN tunnel ACL
                test_config_condition(vid is not None, 'Tunnels do not support VLAN-ACLs')
                # Port & DP tunnel ACL
                test_config_condition(dst_dp_name not in dp_by_name, (
                    f'Could not find referenced destination DP ({dst_dp_name}) for tunnel ACL {acl_in}'))
                dst_dp = dp_by_name[dst_dp_name]
                dst_port = None
                if dst_port_name:
                    dst_port = dst_dp.resolve_port(dst_port_name)
                    test_config_condition(dst_port is None, (
                        f'Could not find referenced destination port ({dst_port_name}) for tunnel ACL {acl_in}'))
                    test_config_condition(dst_port.stack is None, (
                        f'destination port {dst_port_name} for tunnel ACL {acl_in} cannot be a stack port'))
                    dst_port = dst_port.number
                dst_dp = dst_dp.name
                resolved_dst = (dst_dp, dst_port)
                tunnel_vlan = get_tunnel_vlan(tunnel_id_name, resolved_dst)
                # Sources will be resolved later on
                self.tunnel_acls.append(self.acls[acl_in])
                tunnel_dsts_to_vlan[resolved_dst] = tunnel_vlan
                tunnel_id = tunnel_vlan.vid
                return (dst_dp, dst_port, tunnel_id)

            acl.resolve_ports(resolve_port_cb, resolve_tunnel_objects)
            for meter_name in acl.get_meters():
                test_config_condition(meter_name not in self.meters, (
                    f'meter {meter_name} is not configured'))
                acl_meters.add(meter_name)
            for port_no in acl.get_mirror_destinations():
                port = self.ports[port_no]
                port.output_only = True
            return acl.build(self.meters, vid, port_num)

        def verify_acl_exact_match(acls):
            """Verify ACLs have equal exact matches"""
            for acl in acls:
                test_config_condition(acl.exact_match != acls[0].exact_match, (
                    'ACLs when used together must have consistent exact_match'))

        def resolve_acls():
            """Resolve config references in ACLs."""
            for vlan in self.vlans.values():
                if vlan.acls_in:
                    acls = []
                    for acl in vlan.acls_in:
                        resolve_acl(acl, vid=vlan.vid)
                        acls.append(self.acls[acl])
                    vlan.acls_in = acls
                    verify_acl_exact_match(acls)
                if vlan.acls_out:
                    acls = []
                    for acl in vlan.acls_out:
                        resolve_acl(acl, vid=vlan.vid)
                        acls.append(self.acls[acl])
                    vlan.acls_out = acls
                    verify_acl_exact_match(acls)
            for port in self.ports.values():
                if port.acls_in:
                    acls = []
                    test_config_condition(self.dp_acls, (
                        'dataplane ACLs cannot be used with port ACLs.'))
                    for acl in port.acls_in:
                        resolve_acl(acl, port_num=port.number)
                        acls.append(self.acls[acl])
                    port.acls_in = acls
                    verify_acl_exact_match(acls)

                if port.dot1x_dyn_acl:
                    acl_names = [acl_name for acl_name, acl in self.acls.items()
                                 if acl.dot1x_assigned]

                    for acl_name in acl_names:
                        resolve_acl(acl_name, port_num=port.number)

                if port.dot1x_acl:
                    acl_names = [self.dot1x.get('auth_acl'),
                                 self.dot1x.get('noauth_acl')]

                    for acl_name in acl_names:
                        if self.acls.get(acl_name, None):
                            resolve_acl(acl_name, port_num=port.number)

            if self.dp_acls:
                acls = []
                for acl in self.dp_acls:
                    resolve_acl(acl)
                    acls.append(self.acls[acl])
                self.dp_acls = acls
            # Build unbuilt tunnel ACL rules (DP is not the source of the tunnel)
            for acl in self.acls:
                if self.acls[acl].is_tunnel_acl():
                    resolve_acl(acl, None)
            if self.tunnel_acls:
                for tunnel_acl in self.tunnel_acls:
                    tunnel_acl.verify_tunnel_rules()
            self.all_meters = copy.copy(self.meters)
            for unused_meter in set(self.meters.keys()) - acl_meters:
                del self.meters[unused_meter]

        def resolve_routers():
            """Resolve VLAN references in routers."""
            dp_routers = {}
            for router_name, router in self.routers.items():
                if router.bgp_vlan():
                    router.set_bgp_vlan(resolve_vlan(router.bgp_vlan()))
                vlans = resolve_vlans(router.vlans)
                if vlans or router.bgp_vlan():
                    dp_router = copy.copy(router)
                    dp_router.vlans = vlans
                    dp_routers[router_name] = dp_router
            self.routers = dp_routers

            if self.global_vlan:
                vids = {vlan.vid for vlan in self.vlans.values()}
                test_config_condition(
                    self.global_vlan in vids,
                    f'global_vlan VID {self.global_vlan} conflicts with existing VLAN')

            # Check for overlapping VIP subnets or VLANs.
            all_router_vlans = set()
            for router_name, router in self.routers.items():
                vips = set()
                if router.vlans and len(router.vlans) == 1:
                    lone_vlan = router.vlans[0]
                    test_config_condition(
                        lone_vlan in all_router_vlans,
                        f'single VLAN {lone_vlan} in more than one router')
                for vlan in router.vlans:
                    vips.update({vip for vip in vlan.faucet_vips if not vip.ip.is_link_local})
                all_router_vlans.update(router.vlans)
                for vip in vips:
                    for other_vip in vips - set([vip]):
                        test_config_condition(
                            vip.ip in other_vip.network,
                            f'VIPs {vip} and {other_vip} overlap in router {router_name}')
            bgp_routers = self.bgp_routers()
            if bgp_routers:
                for bgp_router in bgp_routers:
                    bgp_vlan = bgp_router.bgp_vlan()
                    vlan_dp_ids = [str(dp.dp_id) for dp in dps if bgp_vlan.vid in dp.vlans]
                    test_config_condition(len(vlan_dp_ids) != 1, (
                        f'DPs ({", ".join(vlan_dp_ids)}) sharing a BGP speaker VLAN ({bgp_vlan.vid}) is unsupported'))
                    test_config_condition(bgp_router.bgp_server_addresses() != (
                        bgp_routers[0].bgp_server_addresses()), (
                            'BGP server addresses must all be the same'))
                router_ids = {bgp_router.bgp_routerid() for bgp_router in bgp_routers}
                test_config_condition(
                    len(router_ids) != 1, f'BGP router IDs must all be the same: {router_ids}')
                bgp_ports = {bgp_router.bgp_port() for bgp_router in bgp_routers}
                test_config_condition(
                    len(bgp_ports) != 1, f'BGP ports must all be the same: {bgp_ports}')

        if not self.stack_ports():
            # Revert back to None if there are no stack ports
            self.stack = None
        if self.stack:
            # Set LLDP defaults for when stacking is configured
            self._lldp_defaults()

        test_config_condition(
            not self.vlans and not self.non_vlan_ports(),
            f'no VLANs referenced by interfaces in {self.name}')
        dp_by_name = {dp.name: dp for dp in dps}
        vlan_by_name = {vlan.name: vlan for vlan in self.vlans.values()}
        loop_protect_external_ports = {
            port for port in self.ports.values() if port.loop_protect_external}
        self.has_externals = bool(loop_protect_external_ports)

        # Populate port.lacp_port_id if it wasn't set in config
        for port in self.ports.values():
            if port.lacp and port.lacp_port_id == -1:
                dp_index = dps.index(self)
                port.lacp_port_id = dp_index * 100 + port.number

        resolve_stack_dps()
        resolve_mirror_destinations()
        resolve_acls()
        resolve_routers()

        for port in self.ports.values():
            port.finalize()
        for vlan in self.vlans.values():
            vlan.finalize()
        for acl in self.acls.values():
            acl.finalize()
        for router in self.routers.values():
            router.finalize()

    def finalize(self):
        """Need to configure OF tables as very last step."""
        self._configure_tables()
        super().finalize()

    def get_native_vlan(self, port_num):
        """Return native VLAN for a port by number, or None."""
        try:
            return self.ports[port_num].native_vlan
        except KeyError:
            return None

    def bgp_routers(self):
        """Return list of routers with BGP enabled."""
        return tuple(
            router for router in self.routers.values() if router.bgp_as() and router.bgp_vlan())

    def dot1x_ports(self):
        """Return list of ports with 802.1x enabled."""
        return tuple(port for port in self.ports.values() if port.dot1x)

    @staticmethod
    def _get_conf_changes(logger, conf_name, subconf, new_subconf, diff=False, ignore_keys=None):
        """Generic detection of config changes between DPs, with merge of unchanged instances."""
        if not ignore_keys:
            ignore_keys = []
        ignore_keys = frozenset(ignore_keys)
        curr_confs = frozenset(subconf.keys())
        new_confs = frozenset(new_subconf.keys())
        deleted_confs = set(curr_confs - new_confs)
        added_confs = set()
        changed_confs = set()
        same_confs = set()
        description_only_confs = set()

        for conf_id, new_conf in new_subconf.items():
            old_conf = subconf.get(conf_id, None)
            if old_conf:
                if old_conf.ignore_subconf(
                        new_conf, ignore_keys=ignore_keys):
                    same_confs.add(conf_id)
                elif old_conf.ignore_subconf(
                        new_conf, ignore_keys=(ignore_keys.union(['description']))):
                    same_confs.add(conf_id)
                    description_only_confs.add(conf_id)
                    logger.info(f'{conf_name} {conf_id} description only changed')
                else:
                    changed_confs.add(conf_id)
                    if diff:
                        logger.info(f'{conf_name} {conf_id} changed: {old_conf.conf_diff(new_conf)}')
                    else:
                        logger.info(f'{conf_name} {conf_id} changed')
            else:
                added_confs.add(conf_id)
                logger.info(f'{conf_name} {conf_id} added')

        for conf_id in same_confs:
            old_conf = subconf[conf_id]
            new_subconf[conf_id].merge_dyn(old_conf)

        changes = deleted_confs or added_confs or changed_confs
        if changes:
            if deleted_confs:
                logger.info(f'{conf_name}s deleted: {deleted_confs}')
            if added_confs:
                logger.info(f'{conf_name}s added: {added_confs}')
            if changed_confs:
                logger.info(f'{conf_name}s changed: {changed_confs}')
        else:
            logger.info(f'no {conf_name} changes')


        return (
            changes, deleted_confs, added_confs, changed_confs, same_confs, description_only_confs)

    def _get_acl_config_changes(self, logger, new_dp):
        """Detect any config changes to ACLs.

        Args:
            logger (ValveLogger): logger instance.
            new_dp (DP): new dataplane configuration.
        Returns:
            changed_acls (set): changed/added ACLs.
        """
        _, _, added_acls, changed_acls, _, _ = self._get_conf_changes(
            logger, 'ACL', self.acls, new_dp.acls, diff=True)
        return added_acls.union(changed_acls)

    def _get_vlan_config_changes(self, logger, new_dp, changed_acls):
        """Detect any config changes to VLANs.

        Args:
            logger (ValveLogger): logger instance.
            new_dp (DP): new dataplane configuration.
            changed_acls (set): changed/added ACL IDs.
        Returns:
            changes (tuple) of:
                deleted_vlans (set): deleted VLAN IDs.
                changed_vlans (set): changed/added VLAN IDs.
        """
        _, deleted_vlans, added_vlans, changed_vlans, same_vlans, _ = self._get_conf_changes(
            logger, 'VLAN', self.vlans, new_dp.vlans,
            diff=True, ignore_keys=frozenset(['acls_in']))
        changed_vlans = added_vlans.union(changed_vlans)
        # TODO: optimize for warm start.
        for vlan_id in same_vlans:
            old_vlan = self.vlans[vlan_id]
            new_vlan = new_dp.vlans[vlan_id]
            if self._acl_ref_changes(
                    'VLAN %u' % vlan_id, old_vlan, new_vlan, changed_acls, logger):
                changed_vlans.add(vlan_id)
        return (deleted_vlans, changed_vlans)

    def _acl_ref_changes(self, conf_desc, old_conf, new_conf, changed_acls, logger):
        changed = False
        new_acl_ids = new_conf.acls_in
        conf_acls_changed = set()
        if new_acl_ids:
            new_acl_ids = [acl._id for acl in new_acl_ids]
            conf_acls_changed = set(new_acl_ids).intersection(changed_acls)
        old_acl_ids = old_conf.acls_in
        if old_acl_ids:
            old_acl_ids = [acl._id for acl in old_acl_ids]
        if conf_acls_changed:
            changed = True
            logger.info('%s ACL changed (ACL %s content changed)' % (
                conf_desc, conf_acls_changed))
        elif (old_acl_ids or new_acl_ids) and old_acl_ids != new_acl_ids:
            changed = True
            logger.info('%s ACL changed (ACL %s to %s)' % (
                conf_desc, old_acl_ids, new_acl_ids))
        return changed

    def _get_port_config_changes(self, logger, new_dp, changed_vlans, deleted_vlans, changed_acls):
        """Detect any config changes to ports.

        Args:
            logger (ValveLogger): logger instance.
            new_dp (DP): new dataplane configuration.
            changed_vlans (set): changed/added VLAN IDs.
            deleted_vlans (set): deleted VLAN IDs.
            changed_acls (set): changed/added ACL IDs.
        Returns:
            changes (tuple) of:
                all_ports_changed (bool): True if all ports changed.
                deleted_ports (set): deleted port numbers.
                changed_ports (set): changed port numbers.
                added_ports (set): added port numbers.
                changed_acl_ports (set): changed ACL only port numbers.
                changed_vlans (set): changed/added VLAN IDs.
        """
        _, deleted_ports, added_ports, changed_ports, same_ports, _ = self._get_conf_changes(
            logger, 'port', self.ports, new_dp.ports,
            diff=True, ignore_keys=frozenset(['acls_in']))

        changed_acl_ports = set()
        all_ports_changed = False

        topology_changed = False
        if self.stack:
            topology_changed = bool(self.stack.hash() != new_dp.stack.hash())
        if topology_changed:
            # Topology changed so restart stack ports just to be safe
            stack_ports = [
                port.number for port in new_dp.stack_ports()
                if port.number not in deleted_ports
                and port.number not in added_ports]
            changed_ports.update(set(stack_ports))
            logger.info('Stack topology change detected, restarting stack ports')
            same_ports -= changed_ports

        if not same_ports:
            all_ports_changed = True
        # TODO: optimize case where only VLAN ACL changed.
        elif changed_vlans:
            all_ports = frozenset(new_dp.ports.keys())
            new_changed_vlans = {
                vlan for vlan in new_dp.vlans.values() if vlan.vid in changed_vlans}
            for vlan in new_changed_vlans:
                changed_port_nums = {port.number for port in vlan.get_ports()}
                changed_ports.update(changed_port_nums)
            all_ports_changed = changed_ports == all_ports

        # Detect changes to VLANs and ACLs based on port changes.
        if not all_ports_changed:
            def get_vids(vlans):
                if not vlans:
                    return set()
                if isinstance(vlans, Iterable):
                    return {vlan.vid for vlan in vlans}
                return {vlans.vid}

            def _add_changed_vlan_port(port, port_dp):
                changed_vlans.update(get_vids(port.vlans()))
                if port.stack:
                    changed_vlans.update(get_vids(port_dp.vlans.values()))

            def _add_changed_vlans(old_port, new_port):
                if old_port.vlans() != new_port.vlans():
                    old_vids = get_vids(old_port.vlans())
                    new_vids = get_vids(new_port.vlans())
                    changed_vlans.update(old_vids.symmetric_difference(new_vids))
                # stacking dis/enabled on a port.
                if bool(old_port.stack) != bool(new_port.stack):
                    changed_vlans.update(get_vids(new_dp.vlans.values()))

            for port_no in changed_ports:
                if port_no not in self.ports:
                    continue
                old_port = self.ports[port_no]
                new_port = new_dp.ports[port_no]
                _add_changed_vlans(old_port, new_port)
            for port_no in deleted_ports:
                port = self.ports[port_no]
                _add_changed_vlan_port(port, self)
            for port_no in added_ports:
                port = new_dp.ports[port_no]
                _add_changed_vlan_port(port, new_dp)
            for port_no in same_ports:
                old_port = self.ports[port_no]
                new_port = new_dp.ports[port_no]
                if old_port.mirror != new_port.mirror:
                    logger.info(f'port {port_no} mirror options changed: {new_port.mirror}')
                    changed_ports.add(port_no)
                if self._acl_ref_changes(
                        'port %u' % port_no, old_port, new_port, changed_acls, logger):
                    changed_acl_ports.add(port_no)

            if changed_acl_ports:
                same_ports -= changed_acl_ports
                logger.info(f'ports where ACL only changed: {changed_acl_ports}')

        same_ports -= changed_ports
        changed_vlans -= deleted_vlans
        # TODO: limit scope to only routers that have affected VLANs.
        changed_vlans_with_vips = []
        for vid in changed_vlans:
            vlan = new_dp.vlans[vid]
            if vlan.faucet_vips:
                changed_vlans_with_vips.append(vlan)
        if changed_vlans_with_vips:
            logger.info(f'forcing cold start because {changed_vlans_with_vips} has routing')
            all_ports_changed = True

        return (all_ports_changed, deleted_ports,
                changed_ports, added_ports, changed_acl_ports,
                changed_vlans)

    def _get_meter_config_changes(self, logger, new_dp):
        """Detect any config changes to meters.
        Args:
            logger (ValveLogger): logger instance.
            new_dp (DP): new dataplane configuration.
        Returns:
            changes (tuple) of:
                deleted_meters (set): deleted Meter IDs.
                changed_meters (set): changed/added Meter IDs.
        """
        (all_meters_changed, deleted_meters,
         added_meters, changed_meters, _, _) = self._get_conf_changes(
             logger, 'METERS', self.meters, new_dp.meters)

        return (all_meters_changed, deleted_meters, added_meters, changed_meters)

    def get_config_changes(self, logger, new_dp):
        """Detect any config changes.

        Args:
            logger (ValveLogger): logger instance
            new_dp (DP): new dataplane configuration.
        Returns:
            (tuple): changes tuple containing:

                deleted_ports (set): deleted port numbers.
                changed_ports (set): changed port numbers.
                added_ports (set): added port numbers.
                changed_acl_ports (set): changed ACL only port numbers.
                deleted_vlans (set): deleted VLAN IDs.
                changed_vlans (set): changed/added VLAN IDs.
                all_ports_changed (bool): True if all ports changed.
                all_meters_changed (bool): True if all meters changed
                deleted_meters (set): deleted meter numbers
                added_meters (set): Added meter numbers
                changed_meters (set): changed/added meter numbers
        """
        if new_dp.stack and self.stack and new_dp.stack.root_name != self.stack.root_name:
            logger.info('Stack root change - requires cold start')
        elif new_dp.routers != self.routers:
            logger.info('DP routers config changed - requires cold start')
        elif not self.ignore_subconf(
                new_dp, ignore_keys=['interfaces', 'interface_ranges', 'routers']):
            logger.info(f'DP config changed - requires cold start: {self.conf_diff(new_dp)}')
        else:
            changed_acls = self._get_acl_config_changes(logger, new_dp)
            deleted_vlans, changed_vlans = self._get_vlan_config_changes(logger, new_dp, changed_acls)
            (all_meters_changed, deleted_meters,
             added_meters, changed_meters) = self._get_meter_config_changes(logger, new_dp)
            (all_ports_changed, deleted_ports, changed_ports, added_ports,
             changed_acl_ports, changed_vlans) = self._get_port_config_changes(
                 logger, new_dp, changed_vlans, deleted_vlans, changed_acls)
            return (deleted_ports, changed_ports, added_ports, changed_acl_ports,
                    deleted_vlans, changed_vlans, all_ports_changed,
                    all_meters_changed, deleted_meters,
                    added_meters, changed_meters)
        # default cold start
        return (set(), set(), set(), set(), set(), set(), True, True, set(), set(), set())

    def get_tables(self):
        """Return tables as dict for API call."""
        return {
            table_name: table.table_id for table_name, table in self.tables.items()}

    def get_config_dict(self):
        """Return DP config as a dict for API call."""
        return {
            'dps': {self.name: self.to_conf()},
            'vlans': {vlan.name: vlan.to_conf() for vlan in self.vlans.values()},
            'acls': {acl_id: acl.to_conf() for acl_id, acl in self.acls.items()}}



#!/usr/bin/env python3

"""Unit tests run as PYTHONPATH=../../.. python3 ./test_valve.py."""

# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest

from os_ken.ofproto import ofproto_v1_3 as ofp

from clib.valve_test_lib import CONFIG, DP1_CONFIG, FAUCET_MAC, ValveTestBases


class ValveTestEgressPipeline(ValveTestBases.ValveTestBig):  # pylint: disable=too-few-public-methods
    """Run complete set of basic tests."""

    DP1_CONFIG = """
            egress_pipeline: True
    """ + DP1_CONFIG


class ValveEgressACLTestCase(ValveTestBases.ValveTestNetwork):
    """Test ACL drop/allow and reloading."""

    def setUp(self):
        self.setup_valves(CONFIG)

    def test_vlan_acl_deny(self):
        """Test VLAN ACL denies a packet."""
        ALLOW_HOST_V6 = 'fc00:200::1:1'  # pylint: disable=invalid-name
        DENY_HOST_V6 = 'fc00:200::1:2'  # pylint: disable=invalid-name
        FAUCET_V100_VIP = 'fc00:100::1'  # pylint: disable=invalid-name
        FAUCET_V200_VIP = 'fc00:200::1'  # pylint: disable=invalid-name
        acl_config = """
dps:
    s1:
{dp1_config}
        interfaces:
            p1:
                number: 1
                native_vlan: v100
            p2:
                number: 2
                native_vlan: v200
                tagged_vlans: [v100]
            p3:
                number: 3
                tagged_vlans: [v100, v200]
            p4:
                number: 4
                tagged_vlans: [v200]
vlans:
    v100:
        vid: 0x100
        faucet_mac: '{mac}'
        faucet_vips: ['{v100_vip}/64']
    v200:
        vid: 0x200
        faucet_mac: '{mac}'
        faucet_vips: ['{v200_vip}/64']
        acl_out: drop_non_allow_host_v6
        minimum_ip_size_check: false
routers:
    r_v100_v200:
        vlans: [v100, v200]
acls:
    drop_non_allow_host_v6:
        - rule:
            ipv6_dst: '{allow_host}'
            eth_type: 0x86DD
            actions:
                allow: 1
        - rule:
            eth_type: 0x86DD
            actions:
                allow: 0
""".format(dp1_config=DP1_CONFIG, mac=FAUCET_MAC, v100_vip=FAUCET_V100_VIP,
           v200_vip=FAUCET_V200_VIP, allow_host=ALLOW_HOST_V6)

        l2_drop_match = {
            'in_port': 2,
            'eth_dst': self.P3_V200_MAC,
            'vlan_vid': 0,
            'eth_type': 0x86DD,
            'ipv6_dst': DENY_HOST_V6}
        l2_accept_match = {
            'in_port': 3,
            'eth_dst': self.P2_V200_MAC,
            'vlan_vid': 0x200 | ofp.OFPVID_PRESENT,
            'eth_type': 0x86DD,
            'ipv6_dst': ALLOW_HOST_V6}
        v100_accept_match = {'in_port': 1, 'vlan_vid': 0}
        table = self.network.tables[self.DP_ID]

        # base case
        for match in (l2_drop_match, l2_accept_match):
            self.assertTrue(
                table.is_output(match, port=4),
                msg='Packet not output before adding ACL')

        # multicast
        self.update_config(acl_config, reload_type='cold')
        self.assertTrue(
            table.is_output(v100_accept_match, port=3),
            msg='Packet not output when on vlan with no ACL'
        )
        self.assertFalse(
            table.is_output(l2_drop_match, port=3),
            msg='Packet not blocked by ACL')
        self.assertTrue(
            table.is_output(l2_accept_match, port=2),
            msg='Packet not allowed by ACL')

        # unicast
        self.rcv_packet(2, 0x200, {
            'eth_src': self.P2_V200_MAC,
            'eth_dst': self.P3_V200_MAC,
            'vid': 0x200,
            'ipv6_src': ALLOW_HOST_V6,
            'ipv6_dst': DENY_HOST_V6,
            'neighbor_advert_ip': ALLOW_HOST_V6,
        })
        self.rcv_packet(3, 0x200, {
            'eth_src': self.P3_V200_MAC,
            'eth_dst': self.P2_V200_MAC,
            'vid': 0x200,
            'ipv6_src': DENY_HOST_V6,
            'ipv6_dst': ALLOW_HOST_V6,
            'neighbor_advert_ip': DENY_HOST_V6,
        })

        self.assertTrue(
            table.is_output(l2_accept_match, port=2),
            msg='Packet not allowed by ACL')
        self.assertFalse(
            table.is_output(l2_drop_match, port=3),
            msg='Packet not blocked by ACL')

        # l3
        l3_drop_match = {
            'in_port': 1,
            'eth_dst': FAUCET_MAC,
            'vlan_vid': 0,
            'eth_type': 0x86DD,
            'ipv6_dst': DENY_HOST_V6}
        l3_accept_match = {
            'in_port': 1,
            'eth_dst': FAUCET_MAC,
            'vlan_vid': 0,
            'eth_type': 0x86DD,
            'ipv6_dst': ALLOW_HOST_V6}

        self.assertTrue(
            table.is_output(l3_accept_match, port=2),
            msg='Routed packet not allowed by ACL')
        self.assertFalse(
            table.is_output(l3_drop_match, port=3),
            msg='Routed packet not blocked by ACL')


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr



#!/usr/bin/env python3

"""Faucet setup script"""

from __future__ import print_function

import errno
import os
import shutil
import sys

from pkg_resources import resource_filename
from setuptools import setup

if sys.version_info < (3,):
    print("""You are trying to install faucet on python {py}

Faucet is not compatible with python 2, please upgrade to python 3.7 or newer."""
          .format(py='.'.join([str(v) for v in sys.version_info[:3]])), file=sys.stderr)
    sys.exit(1)
elif sys.version_info < (3, 7):
    print("""You are trying to install faucet on python {py}

Faucet 1.9.0 and above are no longer compatible with older versions of python 3.

Please upgrade to python 3.7 or newer."""
          .format(py='.'.join([str(v) for v in sys.version_info[:3]])))
    sys.exit(1)


def install_configs():
    """ Install configuration files to /etc """

    dst_ryu_conf_dir = '/etc/faucet/'
    dst_ryu_conf = os.path.join(dst_ryu_conf_dir, 'ryu.conf')
    dst_faucet_conf_dir = '/etc/faucet/'
    src_ryu_conf = resource_filename(__name__, "etc/faucet/ryu.conf")
    src_faucet_conf_dir = resource_filename(__name__, "etc/faucet/")
    faucet_log_dir = '/var/log/faucet/'

    old_ryu_conf = '/etc/ryu/ryu.conf'
    old_faucet_conf_dir = '/etc/ryu/faucet/'

    def setup_ryu_conf():
        if not os.path.exists(dst_ryu_conf_dir):
            print(f"Creating {dst_ryu_conf_dir}")
            os.makedirs(dst_ryu_conf_dir)
        if not os.path.isfile(dst_ryu_conf):
            if os.path.exists(old_ryu_conf) and os.path.isfile(old_ryu_conf):
                print(f"Migrating {old_ryu_conf} to {dst_ryu_conf}")
                shutil.copy(old_ryu_conf, dst_ryu_conf)
            else:
                print(f"Copying {src_ryu_conf} to {dst_ryu_conf}")
                shutil.copy(src_ryu_conf, dst_ryu_conf)

    def setup_faucet_conf():
        if not os.path.exists(dst_faucet_conf_dir):
            print(f"Creating {dst_faucet_conf_dir}")
            os.makedirs(dst_faucet_conf_dir)
        for file_name in os.listdir(src_faucet_conf_dir):
            src_file = os.path.join(src_faucet_conf_dir, file_name)
            dst_file = os.path.join(dst_faucet_conf_dir, file_name)
            alt_src = os.path.join(old_faucet_conf_dir, file_name)
            if os.path.isfile(dst_file):
                continue
            if os.path.isfile(alt_src):
                print(f"Migrating {alt_src} to {dst_file}")
                shutil.copy(alt_src, dst_file)
            elif os.path.isfile(src_file):
                print(f"Copying {src_file} to {dst_file}")
                shutil.copy(src_file, dst_file)

    def setup_faucet_log():
        if not os.path.exists(faucet_log_dir):
            print(f"Creating {faucet_log_dir}")
            os.makedirs(faucet_log_dir)

    try:
        setup_ryu_conf()
        setup_faucet_conf()
    except FileNotFoundError as exception:
        print(str(exception))
    except OSError as exception:
        if exception.errno == errno.EACCES:
            print(f"Permission denied creating {exception.filename}, skipping copying configs")
        else:
            raise

    try:
        setup_faucet_log()
    except OSError as exception:
        print(str(exception))


setup(
    name='faucet',
    setup_requires=['pbr>=1.9', 'setuptools>=17.1'],
    python_requires='>=3.7',
    pbr=True
)

if 'install' in sys.argv or 'bdist_wheel' in sys.argv:
    if os.getenv("DEBINSTALL") is None or (os.getenv("DEBINSTALL") is not None and int(os.environ['DEBINSTALL']) < 1):
        install_configs()



#!/usr/bin/env python3

"""Test FAUCET valve_of."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Innovation Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from faucet import valve_of


class ValveOfTestCase(unittest.TestCase):  # pytype: disable=module-attr
    """Test valve_of functions."""

    def test_reorder_dupe(self):
        """Test simple reordering discards duplicate."""
        flow = valve_of.output_port(1)
        flows = [flow, flow, flow]
        reordered = valve_of.valve_flowreorder(flows, use_barriers=False)
        self.assertEqual(1, len(reordered))

    def test_delete_order(self):
        """Test delete ordering/deupdlication."""
        global_groupdel = valve_of.groupdel(group_id=valve_of.ofp.OFPG_ALL)
        global_flowdel = valve_of.flowmod(
            cookie=None, hard_timeout=None, idle_timeout=None, match_fields=None, out_port=None,
            table_id=valve_of.ofp.OFPTT_ALL, inst=(), priority=0, command=valve_of.ofp.OFPFC_DELETE,
            out_group=valve_of.ofp.OFPG_ANY)
        flowdel = valve_of.flowmod(
            cookie=None, hard_timeout=None, idle_timeout=None, match_fields=None, out_port=None,
            table_id=9, inst=(), priority=0, command=valve_of.ofp.OFPFC_DELETE,
            out_group=valve_of.ofp.OFPG_ANY)
        flow = valve_of.output_port(1)
        flows = [flowdel, flow, flow, flow, global_flowdel, global_groupdel]
        reordered = valve_of.valve_flowreorder(flows, use_barriers=True)
        reordered_str = [str(r) for r in reordered]
        # global deletes come first
        self.assertTrue(valve_of.is_global_groupdel(reordered[0]), msg=reordered)
        self.assertTrue(valve_of.is_global_flowdel(reordered[1]), msg=reordered)
        # with a berrier
        self.assertEqual(str(valve_of.barrier()), str(reordered[2]), msg=reordered)
        # without the individual delete
        self.assertTrue(str(flowdel) not in reordered_str, msg=reordered)
        # with regular flow last
        self.assertEqual(str(flow), reordered_str[-1], msg=reordered)


if __name__ == "__main__":
    unittest.main()  # pytype: disable=module-attr



"""Configure switch tables with TFM messages."""

from faucet import valve_of

REQUIRED_PROPERTIES = set([
    valve_of.ofp.OFPTFPT_WRITE_ACTIONS,
    valve_of.ofp.OFPTFPT_WRITE_ACTIONS_MISS,
    valve_of.ofp.OFPTFPT_APPLY_ACTIONS,
    valve_of.ofp.OFPTFPT_APPLY_ACTIONS_MISS,
    valve_of.ofp.OFPTFPT_WRITE_SETFIELD,
    valve_of.ofp.OFPTFPT_WRITE_SETFIELD_MISS,
    valve_of.ofp.OFPTFPT_MATCH,
    valve_of.ofp.OFPTFPT_WILDCARDS,
    valve_of.ofp.OFPTFPT_APPLY_SETFIELD_MISS,
    valve_of.ofp.OFPTFPT_APPLY_SETFIELD,
    valve_of.ofp.OFPTFPT_NEXT_TABLES,
    valve_of.ofp.OFPTFPT_NEXT_TABLES_MISS,
    valve_of.ofp.OFPTFPT_APPLY_SETFIELD,
    valve_of.ofp.OFPTFPT_INSTRUCTIONS,
    valve_of.ofp.OFPTFPT_INSTRUCTIONS_MISS])


def fill_required_properties(new_table):
    """Ensure TFM has all required properties."""
    configured_props = {prop.type for prop in new_table.properties}
    missing_props = REQUIRED_PROPERTIES - configured_props
    for prop in missing_props:
        new_table.properties.append(
            valve_of.parser.OFPTableFeaturePropOxm(type_=prop))


def init_table(table_id, name, max_entries, metadata_match, metadata_write):
    """Initialize a TFM."""
    if not metadata_match:
        metadata_match = 0
    if not metadata_write:
        metadata_write = 0
    table_attr = {
        'config': 3,
        'max_entries': max_entries,
        'metadata_match': metadata_match,
        'metadata_write': metadata_write,
        'name': name.encode('utf-8'),
        'properties': [],
        'table_id': table_id,
    }
    return valve_of.parser.OFPTableFeaturesStats(**table_attr)


# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def load_tables(dp, valve_cl, max_table_id, min_max_flows, use_oxm_ids, fill_req):
    """Configure switch tables with TFM messages."""
    table_array = []
    active_table_ids = sorted([valve_table.table_id for valve_table in dp.tables.values()])
    for table_id in active_table_ids:
        valve_table = dp.table_by_id(table_id)
        max_entries = max(min_max_flows, valve_table.table_config.size)
        new_table = init_table(
            table_id, valve_table.name, max_entries,
            valve_table.metadata_match, valve_table.metadata_write)
        # Match types
        if valve_table.match_types:
            oxm_ids = []
            if use_oxm_ids:
                oxm_ids = [
                    valve_of.parser.OFPOxmId(type_=match_type, hasmask=hasmask)
                    for match_type, hasmask in valve_table.match_types.items()]
            new_table.properties.append(
                valve_of.parser.OFPTableFeaturePropOxm(
                    oxm_ids=oxm_ids, type_=valve_of.ofp.OFPTFPT_MATCH))
            # Not an exact match table, assume all fields wildcarded.
            if not valve_table.exact_match:
                new_table.properties.append(
                    valve_of.parser.OFPTableFeaturePropOxm(
                        oxm_ids=oxm_ids, type_=valve_of.ofp.OFPTFPT_WILDCARDS))
        insts = set([valve_of.ofp.OFPIT_APPLY_ACTIONS])
        # Next tables
        if valve_table.next_tables:
            new_table.properties.append(valve_of.parser.OFPTableFeaturePropNextTables(
                table_ids=valve_table.next_tables,
                type_=valve_of.ofp.OFPTFPT_NEXT_TABLES))
            insts.add(valve_of.ofp.OFPIT_GOTO_TABLE)
        # Instructions
        if valve_table.table_config.meter:
            insts.add(valve_of.ofp.OFPIT_METER)
        inst_ids = [valve_of.parser.OFPInstructionId(type_) for type_ in insts]
        new_table.properties.append(
            valve_of.parser.OFPTableFeaturePropInstructions(
                type_=valve_of.ofp.OFPTFPT_INSTRUCTIONS, instruction_ids=inst_ids))
        apply_actions = set()
        if valve_table.table_config.dec_ttl and valve_cl.DEC_TTL:
            apply_actions.add(valve_of.ofp.OFPAT_DEC_NW_TTL)
        # Set fields and apply actions
        if valve_table.set_fields:
            apply_actions.add(valve_of.ofp.OFPAT_SET_FIELD)
            if 'vlan_vid' in valve_table.set_fields:
                apply_actions.add(valve_of.ofp.OFPAT_PUSH_VLAN)
            oxm_ids = []
            if use_oxm_ids:
                oxm_ids = [
                    valve_of.parser.OFPOxmId(type_=field, hasmask=False)
                    for field in valve_table.set_fields]
            new_table.properties.append(
                valve_of.parser.OFPTableFeaturePropOxm(
                    oxm_ids=oxm_ids, type_=valve_of.ofp.OFPTFPT_APPLY_SETFIELD))
        if valve_table.table_config.output:
            apply_actions.add(valve_of.ofp.OFPAT_OUTPUT)
            apply_actions.add(valve_of.ofp.OFPAT_POP_VLAN)
            if valve_cl.GROUPS:
                apply_actions.add(valve_of.ofp.OFPAT_GROUP)
        if apply_actions:
            action_ids = [
                valve_of.parser.OFPActionId(type_) for type_ in apply_actions]
            new_table.properties.append(
                valve_of.parser.OFPTableFeaturePropActions(
                    type_=valve_of.ofp.OFPTFPT_APPLY_ACTIONS, action_ids=action_ids))
        # Miss goto table option.
        if valve_table.table_config.miss_goto:
            miss_table_id = dp.tables[valve_table.table_config.miss_goto].table_id
            new_table.properties.append(
                valve_of.parser.OFPTableFeaturePropNextTables(
                    table_ids=[miss_table_id], type_=valve_of.ofp.OFPTFPT_NEXT_TABLES_MISS))
            inst_ids = [valve_of.parser.OFPInstructionId(valve_of.ofp.OFPIT_GOTO_TABLE)]
            new_table.properties.append(
                valve_of.parser.OFPTableFeaturePropInstructions(
                    type_=valve_of.ofp.OFPTFPT_INSTRUCTIONS_MISS, instruction_ids=inst_ids))
        if fill_req:
            fill_required_properties(new_table)
        table_array.append(new_table)

    tfm_table_ids = {table.table_id for table in table_array}
    for missing_table_id in set(range(max_table_id + 1)) - tfm_table_ids:
        new_table = init_table(
            missing_table_id, str(missing_table_id), min_max_flows, 0, 0)
        if fill_req:
            fill_required_properties(new_table)
        table_array.append(new_table)

    return table_array



"""Utility functions for FAUCET."""

# Copyright (C) 2015 Brad Cowie, Christopher Lorier and Joe Stringer.
# Copyright (C) 2015 Research and Education Advanced Network New Zealand Ltd.
# Copyright (C) 2015--2019 The Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from logging.handlers import WatchedFileHandler
import os
import signal
import sys
from functools import wraps

LRU_MAX = 4096


def kill_on_exception(logname):
    """decorator to ensure functions will kill ryu when an unhandled exception
    occurs"""
    def _koe(func):
        @wraps(func)
        def __koe(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except Exception:  # pylint: disable=broad-except
                logging.getLogger(logname).exception('Unhandled exception, killing RYU')
                logging.shutdown()
                os.kill(os.getpid(), signal.SIGTERM)
        return __koe
    return _koe


def utf8_decode(msg_str):
    """Gracefully decode a possibly UTF-8 string."""
    return msg_str.decode('utf-8', errors='replace')


def get_sys_prefix():
    """Returns an additional prefix for log and configuration files when used in
    a virtual environment"""

    # Find the appropriate prefix for config and log file default locations
    # in case Faucet is run in a virtual environment. virtualenv marks the
    # original path in sys.real_prefix. If this value exists, and is
    # different from sys.prefix, then we are most likely running in a
    # virtualenv. Also check for Py3.3+ pyvenv.
    sysprefix = ''
    if (getattr(sys, 'real_prefix', sys.prefix) != sys.prefix
            or getattr(sys, 'base_prefix', sys.prefix) != sys.prefix):
        sysprefix = sys.prefix

    return sysprefix


_PREFIX = get_sys_prefix()
# To specify a boolean-only setting, set the default value to a bool type.
DEFAULTS = {
    'FAUCET_CONFIG': ''.join((
        _PREFIX,
        '/etc/faucet/faucet.yaml',
        ':',
        _PREFIX,
        '/etc/ryu/faucet/faucet.yaml')),
    'FAUCET_STACK_ROOT_STATE_UPDATE_TIME': 10,
    'FAUCET_CONFIG_STAT_RELOAD': False,
    'FAUCET_CONFIG_AUTO_REVERT': False,
    'FAUCET_LOG_LEVEL': 'INFO',
    'FAUCET_LOG': _PREFIX + '/var/log/faucet/faucet.log',
    'FAUCET_EVENT_SOCK': '',  # Special-case, see get_setting().
    'FAUCET_EVENT_SOCK_HEARTBEAT': 0,  # Special-case, see get_setting().
    'FAUCET_EXCEPTION_LOG': _PREFIX + '/var/log/faucet/faucet_exception.log',
    'FAUCET_PROMETHEUS_PORT': '9302',
    'FAUCET_PROMETHEUS_ADDR': '0.0.0.0',
    'GAUGE_CONFIG': ''.join((
        _PREFIX,
        '/etc/faucet/gauge.yaml',
        ':',
        _PREFIX,
        '/etc/ryu/faucet/gauge.yaml')),
    'GAUGE_CONFIG_STAT_RELOAD': False,
    'GAUGE_LOG_LEVEL': 'INFO',
    'GAUGE_PROMETHEUS_ADDR': '0.0.0.0',
    'GAUGE_EXCEPTION_LOG': _PREFIX + '/var/log/faucet/gauge_exception.log',
    'GAUGE_LOG': _PREFIX + '/var/log/faucet/gauge.log'
}


def _cast_bool(value):
    """Return True if value is a non-zero int."""
    try:
        return int(value) != 0
    except ValueError:
        return False


def get_setting(name, path_eval=False):
    """Returns value of specified configuration setting."""
    default_value = DEFAULTS[name]
    result = os.getenv(name, default_value)
    # split on ':' and find the first suitable path
    if (path_eval
            and isinstance(result, str)
            and isinstance(default_value, str) and not
            isinstance(default_value, bool)):
        locations = result.split(":")
        result = None
        for loc in locations:
            if os.path.isfile(loc):
                result = loc
                break
        if result is None:
            result = locations[0]
    # Check for setting that expects a boolean result.
    if isinstance(default_value, bool):
        return _cast_bool(result)
    # Special default for FAUCET_EVENT_SOCK.
    if name == 'FAUCET_EVENT_SOCK':
        if result == '0':
            return ''
        if _cast_bool(result):
            return _PREFIX + '/var/run/faucet/faucet.sock'
    if name == 'FAUCET_EVENT_SOCK_HEARTBEAT':
        if result == '0':
            return 0
    return result


def get_logger(logname, logfile, loglevel, propagate):
    """Create and return a logger object."""

    stream_handlers = {
        'STDOUT': sys.stdout,
        'STDERR': sys.stderr,
    }

    try:
        if logfile in stream_handlers:
            logger_handler = logging.StreamHandler(stream_handlers[logfile])
        else:
            logger_handler = WatchedFileHandler(logfile)
    except (PermissionError, FileNotFoundError) as err:  # pytype: disable=name-error
        print(err)
        sys.exit(-1)

    logger = logging.getLogger(logname)
    log_fmt = '%(asctime)s %(name)-6s %(levelname)-8s %(message)s'
    logger_handler.setFormatter(
        logging.Formatter(log_fmt, '%b %d %H:%M:%S'))
    logger.addHandler(logger_handler)
    logger.propagate = propagate
    logger.setLevel(loglevel)
    return logger


def close_logger(logger):
    """Close all handlers on logger object."""
    if logger is None:
        return
    for handler in list(logger.handlers):
        handler.close()
        logger.removeHandler(handler)


def dpid_log(dpid):
    """Log a DP ID as hex/decimal."""
    if dpid is None:
        return 'DPID None (NoneType)'
    return 'DPID %u (0x%x)' % (dpid, dpid)


def stat_config_files(config_hashes):
    """Return dict of a subset of stat attributes on config files."""
    config_files_stats = {}
    for config_file in list(config_hashes.keys()):
        try:
            config_file_stat = os.stat(config_file)
        except OSError:
            continue
        config_files_stats[config_file] = (
            config_file_stat.st_size,
            config_file_stat.st_mtime,
            config_file_stat.st_ctime)
    return config_files_stats
